import re
import urllib.parse
import warnings
from abc import ABC
from collections import Counter, defaultdict
from collections.abc import Collection, Mapping
from datetime import date, datetime
from functools import cached_property
from typing import Any, ClassVar, Literal, TypeVar, cast, overload

from cognite.client.data_classes import data_modeling as dms
from cognite.client.data_classes.data_modeling import DataModelId, DataModelIdentifier, ViewId
from cognite.client.utils.useful_types import SequenceNotStr
from pydantic import ValidationError
from rdflib import Namespace

from cognite.neat._client import NeatClient
from cognite.neat._client.data_classes.data_modeling import ContainerApplyDict, ViewApplyDict
from cognite.neat._constants import (
    COGNITE_MODELS,
    COGNITE_SPACES,
    DMS_CONTAINER_PROPERTY_SIZE_LIMIT,
    DMS_RESERVED_PROPERTIES,
    get_default_prefixes_and_namespaces,
)
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings import NeatValueWarning
from cognite.neat._issues.warnings._models import (
    SolutionModelBuildOnTopOfCDMWarning,
)
from cognite.neat._rules._constants import PATTERNS, get_reserved_words
from cognite.neat._rules._shared import (
    ReadInputRules,
    ReadRules,
    VerifiedRules,
)
from cognite.neat._rules.analysis import RulesAnalysis
from cognite.neat._rules.importers import DMSImporter
from cognite.neat._rules.models import (
    DMSInputRules,
    DMSRules,
    InformationInputRules,
    InformationRules,
    SheetList,
    data_types,
)
from cognite.neat._rules.models.data_types import AnyURI, DataType, Enum, File, String, Timeseries
from cognite.neat._rules.models.dms import DMSMetadata, DMSProperty, DMSValidation, DMSView
from cognite.neat._rules.models.dms._rules import DMSContainer, DMSEnum, DMSNode
from cognite.neat._rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    DMSUnknownEntity,
    EdgeEntity,
    HasDataFilter,
    MultiValueTypeInfo,
    ReverseConnectionEntity,
    UnknownEntity,
    ViewEntity,
)
from cognite.neat._rules.models.information import InformationClass, InformationMetadata, InformationProperty
from cognite.neat._utils.rdf_ import get_inheritance_path
from cognite.neat._utils.text import NamingStandardization, title, to_camel_case, to_words

from ._base import RulesTransformer, T_VerifiedIn, T_VerifiedOut, VerifiedRulesTransformer
from ._verification import VerifyDMSRules

T_InputInRules = TypeVar("T_InputInRules", bound=ReadInputRules)
T_InputOutRules = TypeVar("T_InputOutRules", bound=ReadInputRules)


class ConversionTransformer(VerifiedRulesTransformer[T_VerifiedIn, T_VerifiedOut], ABC):
    """Base class for all conversion transformers."""

    ...


class ToDMSCompliantEntities(RulesTransformer[ReadRules[InformationInputRules], ReadRules[InformationInputRules]]):
    """Converts input rules to rules that is compliant with the Information Model.

    This is typically used with importers from arbitrary sources to ensure that classes and properties have valid
    names.

    Args:
        rename_warning: How to handle renaming of entities that are not compliant with the Information Model.
            - "raise": Raises a warning and renames the entity.
            - "skip": Renames the entity without raising a warning.
    """

    def __init__(self, rename_warning: Literal["raise", "skip"] = "skip") -> None:
        self._renaming = rename_warning

    @property
    def description(self) -> str:
        return "Ensures that all entities are compliant with the Information Model."

    def transform(self, rules: ReadRules[InformationInputRules]) -> ReadRules[InformationInputRules]:
        if rules.rules is None:
            return rules
        # Doing dump to obtain a copy, and ensure that all entities are created. Input allows
        # string for entities, the dump call will convert these to entities.
        dumped = rules.rules.dump()
        copy = InformationInputRules.load(dumped)

        new_by_old_class_suffix: dict[str, str] = {}
        for cls in copy.classes:
            cls_entity = cast(ClassEntity, cls.class_)  # Safe due to the dump above
            if not PATTERNS.view_id_compliance.match(cls_entity.suffix):
                new_suffix = self._fix_cls_suffix(cls_entity.suffix)
                if self._renaming == "raise":
                    warnings.warn(
                        NeatValueWarning(f"Invalid class name {cls_entity.suffix!r}.Renaming to {new_suffix}"),
                        stacklevel=2,
                    )
                cls.class_.suffix = new_suffix  # type: ignore[union-attr]

        for cls_ in copy.classes:
            if cls_.implements:
                for i, parent in enumerate(cls_.implements):
                    if isinstance(parent, ClassEntity) and parent.suffix in new_by_old_class_suffix:
                        cls_.implements[i].suffix = new_by_old_class_suffix[parent.suffix]  # type: ignore[union-attr]

        for prop in copy.properties:
            if not PATTERNS.dms_property_id_compliance.match(prop.property_):
                new_property = self._fix_property(prop.property_)
                if self._renaming == "warning":
                    warnings.warn(
                        NeatValueWarning(
                            f"Invalid property name {prop.class_.suffix}.{prop.property_!r}. Renaming to {new_property}"  # type: ignore[union-attr]
                        ),
                        stacklevel=2,
                    )
                prop.property_ = new_property

            if isinstance(prop.class_, ClassEntity) and prop.class_.suffix in new_by_old_class_suffix:
                prop.class_.suffix = new_by_old_class_suffix[prop.class_.suffix]

            if isinstance(prop.value_type, ClassEntity) and prop.value_type.suffix in new_by_old_class_suffix:
                prop.value_type.suffix = new_by_old_class_suffix[prop.value_type.suffix]

            if isinstance(prop.value_type, MultiValueTypeInfo):
                for i, value_type in enumerate(prop.value_type.types):
                    if isinstance(value_type, ClassEntity) and value_type.suffix in new_by_old_class_suffix:
                        prop.value_type.types[i].suffix = new_by_old_class_suffix[value_type.suffix]  # type: ignore[union-attr]

        return ReadRules(rules=copy, read_context=rules.read_context)

    @cached_property
    def _reserved_class_words(self) -> set[str]:
        return set(get_reserved_words("class"))

    @cached_property
    def _reserved_property_words(self) -> set[str]:
        return set(get_reserved_words("property"))

    def _fix_cls_suffix(self, suffix: str) -> str:
        if suffix in self._reserved_class_words:
            return f"My{suffix}"
        suffix = urllib.parse.unquote(suffix)
        suffix = NamingStandardization.standardize_class_str(suffix)
        if len(suffix) > 252:
            suffix = suffix[:252]
        return suffix

    def _fix_property(self, property_: str) -> str:
        if property_ in self._reserved_property_words:
            return f"my{property_}"
        property_ = urllib.parse.unquote(property_)
        property_ = NamingStandardization.standardize_property_str(property_)
        if len(property_) > 252:
            property_ = property_[:252]
        return property_


class StandardizeSpaceAndVersion(VerifiedRulesTransformer[DMSRules, DMSRules]):  # type: ignore[misc]
    """This transformer standardizes the space and version of the DMSRules.

    typically used to ensure all the views are moved to the same version as the data model.

    """

    @property
    def description(self) -> str:
        return "Ensures uniform version and space of the views belonging to the data model."

    def transform(self, rules: DMSRules) -> DMSRules:
        copy = rules.model_copy(deep=True)

        space = copy.metadata.space
        version = copy.metadata.version

        copy.views = self._standardize_views(copy.views, space, version)
        copy.properties = self._standardize_properties(copy.properties, space, version)
        return copy

    def _standardize_views(self, views: SheetList[DMSView], space: str, version: str) -> SheetList[DMSView]:
        for view in views:
            if view.view.space not in COGNITE_SPACES:
                view.view.version = version
                view.view.prefix = space

            if view.implements:
                for i, parent in enumerate(view.implements):
                    if parent.space not in COGNITE_SPACES:
                        view.implements[i].version = version
                        view.implements[i].prefix = space
        return views

    def _standardize_properties(
        self, properties: SheetList[DMSProperty], space: str, version: str
    ) -> SheetList[DMSProperty]:
        for property_ in properties:
            if property_.view.space not in COGNITE_SPACES:
                property_.view.version = version
                property_.view.prefix = space

            if isinstance(property_.value_type, ViewEntity) and property_.value_type.space not in COGNITE_SPACES:
                property_.value_type.version = version
                property_.value_type.prefix = space

            # for edge connection
            if (
                property_.connection
                and isinstance(property_.connection, EdgeEntity)
                and property_.connection.properties
            ):
                if property_.connection.properties.space not in COGNITE_SPACES:
                    property_.connection.properties.version = version
                    property_.connection.properties.prefix = space

        return properties


class ToCompliantEntities(VerifiedRulesTransformer[InformationRules, InformationRules]):  # type: ignore[misc]
    """Converts input rules to rules with compliant entity IDs that match regex patters used
    by DMS schema components."""

    @property
    def description(self) -> str:
        return "Ensures externalIDs are compliant with CDF"

    def transform(self, rules: InformationRules) -> InformationRules:
        copy = rules.model_copy(deep=True)
        copy.classes = self._fix_classes(copy.classes)
        copy.properties = self._fix_properties(copy.properties)
        return copy

    @classmethod
    def _fix_entity(cls, entity: str) -> str:
        entity = re.sub(r"[^_a-zA-Z0-9]+", "_", entity)

        # entity id must start with a letter
        if not entity[0].isalpha():
            entity = "prefix_" + entity
        # and end with a letter or number
        if not entity[-1].isalnum():
            entity = entity + "_suffix"

        # removing any double underscores that could occur
        return re.sub(r"[^a-zA-Z0-9]+", "_", entity)

    @classmethod
    def _fix_class(cls, class_: ClassEntity) -> ClassEntity:
        if isinstance(class_, ClassEntity) and type(class_.prefix) is str:
            class_ = ClassEntity(
                prefix=cls._fix_entity(class_.prefix),
                suffix=cls._fix_entity(class_.suffix),
            )

        return class_

    @classmethod
    def _fix_value_type(
        cls, value_type: DataType | ClassEntity | MultiValueTypeInfo
    ) -> DataType | ClassEntity | MultiValueTypeInfo:
        fixed_value_type: DataType | ClassEntity | MultiValueTypeInfo

        # value type specified as MultiValueTypeInfo
        if isinstance(value_type, MultiValueTypeInfo):
            fixed_value_type = MultiValueTypeInfo(
                types=[cast(DataType | ClassEntity, cls._fix_value_type(type_)) for type_ in value_type.types],
            )

        # value type specified as ClassEntity instance
        elif isinstance(value_type, ClassEntity):
            fixed_value_type = cls._fix_class(value_type)

        # this is a DataType instance but also we should default to original value
        else:
            fixed_value_type = value_type

        return fixed_value_type

    @classmethod
    def _fix_classes(cls, definitions: SheetList[InformationClass]) -> SheetList[InformationClass]:
        fixed_definitions = SheetList[InformationClass]()
        for definition in definitions:
            definition.class_ = cls._fix_class(definition.class_)
            fixed_definitions.append(definition)
        return fixed_definitions

    @classmethod
    def _fix_properties(cls, definitions: SheetList[InformationProperty]) -> SheetList[InformationProperty]:
        fixed_definitions = SheetList[InformationProperty]()
        for definition in definitions:
            definition.class_ = cls._fix_class(definition.class_)
            definition.property_ = cls._fix_entity(definition.property_)
            definition.value_type = cls._fix_value_type(definition.value_type)
            fixed_definitions.append(definition)
        return fixed_definitions


class PrefixEntities(ConversionTransformer):  # type: ignore[type-var]
    """Prefixes all entities with a given prefix if they are in the same space as data model."""

    def __init__(self, prefix: str) -> None:
        self._prefix = prefix

    @property
    def description(self) -> str:
        return f"Prefixes all entities with {self._prefix!r} prefix if they are in the same space as data model."

    @overload
    def transform(self, rules: DMSRules) -> DMSRules: ...

    @overload
    def transform(self, rules: InformationRules) -> InformationRules: ...

    def transform(self, rules: InformationRules | DMSRules) -> InformationRules | DMSRules:
        copy: InformationRules | DMSRules = rules.model_copy(deep=True)

        # Case: Prefix Information Rules
        if isinstance(copy, InformationRules):
            # prefix classes
            for cls in copy.classes:
                if cls.class_.prefix == copy.metadata.prefix:
                    cls.class_ = self._with_prefix(cls.class_)

                if cls.implements:
                    # prefix parents
                    for i, parent_class in enumerate(cls.implements):
                        if parent_class.prefix == copy.metadata.prefix:
                            cls.implements[i] = self._with_prefix(parent_class)

            for prop in copy.properties:
                if prop.class_.prefix == copy.metadata.prefix:
                    prop.class_ = self._with_prefix(prop.class_)

                # value type property is not multi and it is ClassEntity

                if isinstance(prop.value_type, ClassEntity) and prop.value_type.prefix == copy.metadata.prefix:
                    prop.value_type = self._with_prefix(cast(ClassEntity, prop.value_type))
                elif isinstance(prop.value_type, MultiValueTypeInfo):
                    for i, value_type in enumerate(prop.value_type.types):
                        if isinstance(value_type, ClassEntity) and value_type.prefix == copy.metadata.prefix:
                            prop.value_type.types[i] = self._with_prefix(cast(ClassEntity, value_type))
            return copy

        # Case: Prefix DMS Rules
        elif isinstance(copy, DMSRules):
            for view in copy.views:
                if view.view.space == copy.metadata.space:
                    view.view = self._with_prefix(view.view)

                if view.implements:
                    for i, parent_view in enumerate(view.implements):
                        if parent_view.space == copy.metadata.space:
                            view.implements[i] = self._with_prefix(parent_view)

            for dms_prop in copy.properties:
                if dms_prop.view.space == copy.metadata.space:
                    dms_prop.view = self._with_prefix(dms_prop.view)

                if isinstance(dms_prop.value_type, ViewEntity) and dms_prop.value_type.space == copy.metadata.space:
                    dms_prop.value_type = self._with_prefix(dms_prop.value_type)

                if isinstance(dms_prop.container, ContainerEntity) and dms_prop.container.space == copy.metadata.space:
                    dms_prop.container = self._with_prefix(dms_prop.container)

            if copy.containers:
                for container in copy.containers:
                    if container.container.space == copy.metadata.space:
                        container.container = self._with_prefix(container.container)
            return copy

        raise NeatValueError(f"Unsupported rules type: {type(copy)}")

    @overload
    def _with_prefix(self, entity: ClassEntity) -> ClassEntity: ...

    @overload
    def _with_prefix(self, entity: ViewEntity) -> ViewEntity: ...

    @overload
    def _with_prefix(self, entity: ContainerEntity) -> ContainerEntity: ...

    def _with_prefix(
        self, entity: ViewEntity | ContainerEntity | ClassEntity
    ) -> ViewEntity | ContainerEntity | ClassEntity:
        if isinstance(entity, ViewEntity | ContainerEntity | ClassEntity):
            entity.suffix = f"{self._prefix}{entity.suffix}"

        else:
            raise NeatValueError(f"Unsupported entity type: {type(entity)}")

        return entity


class StandardizeNaming(ConversionTransformer):
    """Sets views/classes/container names to PascalCase and properties to camelCase."""

    @property
    def description(self) -> str:
        return "Sets views/classes/containers names to PascalCase and properties to camelCase."

    @overload
    def transform(self, rules: DMSRules) -> DMSRules: ...

    @overload
    def transform(self, rules: InformationRules) -> InformationRules: ...

    def transform(self, rules: InformationRules | DMSRules) -> InformationRules | DMSRules:
        output = rules.model_copy(deep=True)
        if isinstance(output, InformationRules):
            return self._standardize_information_rules(output)
        elif isinstance(output, DMSRules):
            return self._standardize_dms_rules(output)
        raise NeatValueError(f"Unsupported rules type: {type(output)}")

    def _standardize_information_rules(self, rules: InformationRules) -> InformationRules:
        new_by_old_class_suffix: dict[str, str] = {}
        for cls in rules.classes:
            new_suffix = NamingStandardization.standardize_class_str(cls.class_.suffix)
            new_by_old_class_suffix[cls.class_.suffix] = new_suffix
            cls.class_.suffix = new_suffix

        for cls in rules.classes:
            if cls.implements:
                for i, parent in enumerate(cls.implements):
                    if parent.suffix in new_by_old_class_suffix:
                        cls.implements[i].suffix = new_by_old_class_suffix[parent.suffix]

        for prop in rules.properties:
            prop.property_ = NamingStandardization.standardize_property_str(prop.property_)
            if prop.class_.suffix in new_by_old_class_suffix:
                prop.class_.suffix = new_by_old_class_suffix[prop.class_.suffix]

            if isinstance(prop.value_type, ClassEntity) and prop.value_type.suffix in new_by_old_class_suffix:
                prop.value_type.suffix = new_by_old_class_suffix[prop.value_type.suffix]

            if isinstance(prop.value_type, MultiValueTypeInfo):
                for i, value_type in enumerate(prop.value_type.types):
                    if isinstance(value_type, ClassEntity) and value_type.suffix in new_by_old_class_suffix:
                        prop.value_type.types[i].suffix = new_by_old_class_suffix[value_type.suffix]  # type: ignore[union-attr]

        return rules

    def _standardize_dms_rules(self, rules: DMSRules) -> DMSRules:
        new_by_old_view: dict[str, str] = {}
        for view in rules.views:
            new_suffix = NamingStandardization.standardize_class_str(view.view.suffix)
            new_by_old_view[view.view.suffix] = new_suffix
            view.view.suffix = new_suffix
        new_by_old_container: dict[str, str] = {}
        if rules.containers:
            for container in rules.containers:
                new_suffix = NamingStandardization.standardize_class_str(container.container.suffix)
                new_by_old_container[container.container.suffix] = new_suffix
                container.container.suffix = new_suffix

        for view in rules.views:
            if view.implements:
                for i, parent in enumerate(view.implements):
                    if parent.suffix in new_by_old_view:
                        view.implements[i].suffix = new_by_old_view[parent.suffix]
            if view.filter_ and isinstance(view.filter_, HasDataFilter) and view.filter_.inner:
                for i, item in enumerate(view.filter_.inner):
                    if isinstance(item, ContainerEntity) and item.suffix in new_by_old_container:
                        view.filter_.inner[i].suffix = new_by_old_container[item.suffix]
                    if isinstance(item, ViewEntity) and item.suffix in new_by_old_view:
                        view.filter_.inner[i].suffix = new_by_old_view[item.suffix]
        if rules.containers:
            for container in rules.containers:
                if container.constraint:
                    for i, constraint in enumerate(container.constraint):
                        if constraint.suffix in new_by_old_container:
                            container.constraint[i].suffix = new_by_old_container[constraint.suffix]
        new_property_by_view_by_old_property: dict[ViewEntity, dict[str, str]] = defaultdict(dict)
        for prop in rules.properties:
            if prop.view.suffix in new_by_old_view:
                prop.view.suffix = new_by_old_view[prop.view.suffix]
            new_view_property = NamingStandardization.standardize_property_str(prop.view_property)
            new_property_by_view_by_old_property[prop.view][prop.view_property] = new_view_property
            prop.view_property = new_view_property
            if isinstance(prop.value_type, ViewEntity) and prop.value_type.suffix in new_by_old_view:
                prop.value_type.suffix = new_by_old_view[prop.value_type.suffix]
            if (
                isinstance(prop.connection, EdgeEntity)
                and prop.connection.properties
                and prop.connection.properties.suffix in new_by_old_view
            ):
                prop.connection.properties.suffix = new_by_old_view[prop.connection.properties.suffix]
            if isinstance(prop.container, ContainerEntity) and prop.container.suffix in new_by_old_container:
                prop.container.suffix = new_by_old_container[prop.container.suffix]
            if prop.container_property:
                prop.container_property = NamingStandardization.standardize_property_str(prop.container_property)
        for prop in rules.properties:
            if (
                isinstance(prop.connection, ReverseConnectionEntity)
                and isinstance(prop.value_type, ViewEntity)
                and prop.value_type in new_property_by_view_by_old_property
            ):
                new_by_old_property = new_property_by_view_by_old_property[prop.value_type]
                if prop.connection.property_ in new_by_old_property:
                    prop.connection.property_ = new_by_old_property[prop.connection.property_]
        return rules


class InformationToDMS(ConversionTransformer[InformationRules, DMSRules]):
    """Converts InformationRules to DMSRules."""

    def __init__(
        self, ignore_undefined_value_types: bool = False, reserved_properties: Literal["error", "warning"] = "error"
    ):
        self.ignore_undefined_value_types = ignore_undefined_value_types
        self.reserved_properties = reserved_properties

    def transform(self, rules: InformationRules) -> DMSRules:
        return _InformationRulesConverter(rules).as_dms_rules(
            self.ignore_undefined_value_types, self.reserved_properties
        )


class DMSToInformation(ConversionTransformer[DMSRules, InformationRules]):
    """Converts DMSRules to InformationRules."""

    def __init__(self, instance_namespace: Namespace | None = None):
        self.instance_namespace = instance_namespace

    def transform(self, rules: DMSRules) -> InformationRules:
        return _DMSRulesConverter(rules, self.instance_namespace).as_information_rules()


class ConvertToRules(ConversionTransformer[VerifiedRules, VerifiedRules]):
    """Converts any rules to any rules."""

    def __init__(self, out_cls: type[VerifiedRules]):
        self._out_cls = out_cls

    def transform(self, rules: VerifiedRules) -> VerifiedRules:
        if isinstance(rules, self._out_cls):
            return rules
        if isinstance(rules, InformationRules) and self._out_cls is DMSRules:
            return InformationToDMS().transform(rules)
        if isinstance(rules, DMSRules) and self._out_cls is InformationRules:
            return DMSToInformation().transform(rules)
        raise ValueError(f"Unsupported conversion from {type(rules)} to {self._out_cls}")


_T_Entity = TypeVar("_T_Entity", bound=ClassEntity | ViewEntity)


class SetIDDMSModel(VerifiedRulesTransformer[DMSRules, DMSRules]):
    def __init__(self, new_id: DataModelId | tuple[str, str, str], name: str | None = None):
        self.new_id = DataModelId.load(new_id)
        self.name = name

    @property
    def description(self) -> str:
        return f"Sets the Data Model ID to {self.new_id.as_tuple()}"

    def transform(self, rules: DMSRules) -> DMSRules:
        if self.new_id.version is None:
            raise NeatValueError("Version is required when setting a new Data Model ID")
        dump = rules.dump()
        dump["metadata"]["space"] = self.new_id.space
        dump["metadata"]["external_id"] = self.new_id.external_id
        dump["metadata"]["version"] = self.new_id.version
        dump["metadata"]["name"] = self.name or self._generate_name()
        # Serialize and deserialize to set the new space and external_id
        # as the default values for the new model.
        return DMSRules.model_validate(DMSInputRules.load(dump).dump())

    def _generate_name(self) -> str:
        return title(to_words(self.new_id.external_id))


class ToExtensionModel(VerifiedRulesTransformer[DMSRules, DMSRules], ABC):
    type_: ClassVar[str]

    def __init__(self, new_model_id: DataModelIdentifier) -> None:
        self.new_model_id = DataModelId.load(new_model_id)
        if not self.new_model_id.version:
            raise NeatValueError("Version is required for the new model.")

    @property
    def description(self) -> str:
        return f"Create new data model {self.new_model_id} of type {self.type_.replace('_', ' ')} data model."


class ToEnterpriseModel(ToExtensionModel):
    type_: ClassVar[str] = "enterprise"

    def __init__(
        self,
        new_model_id: DataModelIdentifier,
        org_name: str = "My",
        dummy_property: str = "GUID",
        move_connections: bool = False,
    ):
        super().__init__(new_model_id)
        self.dummy_property = dummy_property
        self.org_name = org_name
        self.move_connections = move_connections

    def transform(self, rules: DMSRules) -> DMSRules:
        return self._to_enterprise(rules)

    def _to_enterprise(self, reference_model: DMSRules) -> DMSRules:
        enterprise_model = reference_model.model_copy(deep=True)

        enterprise_model.metadata.name = f"{self.org_name} {self.type_} data model"
        enterprise_model.metadata.space = self.new_model_id.space
        enterprise_model.metadata.external_id = self.new_model_id.external_id
        enterprise_model.metadata.version = cast(str, self.new_model_id.version)

        # Here we are creating enterprise views with a single container with a dummy property
        # for each view
        enterprise_views, enterprise_containers, enterprise_properties = self._create_new_views(enterprise_model)

        # We keep the reference views, and adding new enterprise views...
        enterprise_model.views.extend(enterprise_views)

        if self.move_connections:
            # Move connections from reference model to new enterprise model
            enterprise_properties.extend(self._create_connection_properties(enterprise_model, enterprise_views))

        # ... however, we do not want to keep the reference containers and properties
        # these we are getting for free through the implements.
        enterprise_containers.sort(key=lambda container: (container.container.space, container.container.external_id))
        enterprise_model.containers = enterprise_containers
        enterprise_properties.sort(
            key=lambda prop: (prop.view.space, prop.view.external_id, prop.view.version, prop.view_property)
        )
        enterprise_model.properties = enterprise_properties

        # Sorting all your views first.
        enterprise_model.views.sort(
            key=lambda view: (
                # Sorting your views first
                int(view.view.space != self.new_model_id.space),
                view.view.space,
                view.view.external_id,
                view.view.version,
            )
        )
        return enterprise_model

    @staticmethod
    def _create_connection_properties(rules: DMSRules, new_views: SheetList[DMSView]) -> SheetList[DMSProperty]:
        """Creates a new connection property for each connection property in the reference model.

        This is for example when you create an enterprise model from CogniteCore, you ensure that your
        new Asset, Equipment, TimeSeries, Activity, and File views all point to each other.
        """
        # Note all new news have an implements attribute that points to the original view
        previous_by_new_view = {view.implements[0]: view.view for view in new_views if view.implements}
        connection_properties = SheetList[DMSProperty]()
        for prop in rules.properties:
            if (
                isinstance(prop.value_type, ViewEntity)
                and prop.view in previous_by_new_view
                and prop.value_type in previous_by_new_view
            ):
                new_property = prop.model_copy(deep=True)
                new_property.view = previous_by_new_view[prop.view]
                new_property.value_type = previous_by_new_view[prop.value_type]
                connection_properties.append(new_property)

        return connection_properties

    def _create_new_views(
        self, rules: DMSRules
    ) -> tuple[SheetList[DMSView], SheetList[DMSContainer], SheetList[DMSProperty]]:
        """Creates new views for the new model.

        If the dummy property is provided, it will also create a new container for each view
        with a single property that is the dummy property.
        """
        new_views = SheetList[DMSView]()
        new_containers = SheetList[DMSContainer]()
        new_properties = SheetList[DMSProperty]()

        for definition in rules.views:
            view_entity = self._remove_cognite_affix(definition.view)
            view_entity.version = cast(str, self.new_model_id.version)
            view_entity.prefix = self.new_model_id.space
            new_views.append(
                DMSView(
                    view=view_entity,
                    implements=[definition.view],
                    in_model=True,
                    name=definition.name,
                )
            )

            if self.dummy_property is None:
                continue

            container_entity = ContainerEntity(space=view_entity.prefix, externalId=view_entity.external_id)

            container = DMSContainer(container=container_entity)

            property_id = f"{to_camel_case(view_entity.suffix)}{self.dummy_property}"
            property_ = DMSProperty(
                view=view_entity,
                view_property=property_id,
                value_type=String(),
                nullable=True,
                immutable=False,
                is_list=False,
                container=container_entity,
                container_property=property_id,
            )

            new_properties.append(property_)
            new_containers.append(container)

        return new_views, new_containers, new_properties

    def _remove_cognite_affix(self, entity: ViewEntity) -> ViewEntity:
        """This method removes `Cognite` affix from the entity."""
        new_suffix = entity.suffix.replace("Cognite", self.org_name)
        return ViewEntity(space=entity.space, externalId=new_suffix, version=entity.version)


class ToSolutionModel(ToExtensionModel):
    """Creates a solution data model based on an existing data model.

    The solution data model will create a new view for each view in the existing model.

    Args:
        new_model_id: DataData model identifier for the new model.
        dummy_property: Only applicable if mode='write'. The identifier of the dummy property in the newly created
            container.
        exclude_views_in_other_spaces: Whether to exclude views that are not in the same space as the existing model,
            when creating the solution model.

    """

    type_: ClassVar[str] = "solution"

    def __init__(
        self,
        new_model_id: DataModelIdentifier,
        properties: Literal["repeat", "connection"] = "connection",
        dummy_property: str | None = "GUID",
        direct_property: str = "enterprise",
        view_prefix: str = "Enterprise",
        filter_type: Literal["container", "view"] = "container",
        exclude_views_in_other_spaces: bool = False,
        skip_cognite_views: bool = True,
    ):
        super().__init__(new_model_id)
        self.properties = properties
        self.dummy_property = dummy_property
        self.direct_property = direct_property
        self.view_prefix = view_prefix
        self.filter_type = filter_type
        self.exclude_views_in_other_spaces = exclude_views_in_other_spaces
        self.skip_cognite_views = skip_cognite_views

    def transform(self, rules: DMSRules) -> DMSRules:
        reference_model = rules
        reference_model_id = reference_model.metadata.as_data_model_id()
        if reference_model_id in COGNITE_MODELS:
            warnings.warn(
                SolutionModelBuildOnTopOfCDMWarning(reference_model_id=reference_model_id),
                stacklevel=2,
            )
        return self._to_solution(reference_model)

    def _to_solution(self, reference_rules: DMSRules) -> DMSRules:
        """For creation of solution data model / rules specifically for mapping over existing containers."""
        reference_rules = self._expand_properties(reference_rules.model_copy(deep=True))

        new_views, new_properties, read_view_by_new_view = self._create_views(reference_rules)
        new_containers, new_container_properties = self._create_containers_update_view_filter(
            new_views, reference_rules, read_view_by_new_view
        )
        new_properties.extend(new_container_properties)

        if self.properties == "connection":
            # Ensure the Enterprise view and the new solution view are next to each other
            new_properties.sort(
                key=lambda prop: (
                    prop.view.external_id.removeprefix(self.view_prefix),
                    bool(prop.view.external_id.startswith(self.view_prefix)),
                )
            )
        else:
            new_properties.sort(key=lambda prop: (prop.view.external_id, prop.view_property))

        metadata = reference_rules.metadata.model_copy(
            deep=True,
            update={
                "space": self.new_model_id.space,
                "external_id": self.new_model_id.external_id,
                "version": self.new_model_id.version,
                "name": f"{self.type_} data model",
            },
        )
        return DMSRules(
            metadata=metadata,
            properties=new_properties,
            views=new_views,
            containers=new_containers or None,
            enum=reference_rules.enum,
            nodes=reference_rules.nodes,
        )

    @staticmethod
    def _expand_properties(rules: DMSRules) -> DMSRules:
        probe = RulesAnalysis(dms=rules)
        ancestor_properties_by_view = probe.properties_by_view(
            include_ancestors=True,
            include_different_space=True,
        )
        property_ids_by_view = {
            view: {prop.view_property for prop in properties}
            for view, properties in probe.properties_by_view(
                include_ancestors=False, include_different_space=True
            ).items()
        }
        for view, property_ids in property_ids_by_view.items():
            ancestor_properties = ancestor_properties_by_view.get(view, [])
            for prop in ancestor_properties:
                if isinstance(prop.connection, ReverseConnectionEntity):
                    # If you try to add a reverse direct relation of a parent, it will fail as the ValueType of the
                    # original property will point to the parent view, and not the child.
                    continue
                if prop.view_property not in property_ids:
                    rules.properties.append(prop)
                    property_ids.add(prop.view_property)
        return rules

    def _create_views(
        self, reference: DMSRules
    ) -> tuple[SheetList[DMSView], SheetList[DMSProperty], dict[ViewEntity, ViewEntity]]:
        renaming: dict[ViewEntity, ViewEntity] = {}
        new_views = SheetList[DMSView]()
        read_view_by_new_view: dict[ViewEntity, ViewEntity] = {}
        for ref_view in reference.views:
            if (self.skip_cognite_views and ref_view.view.space in COGNITE_SPACES) or (
                self.exclude_views_in_other_spaces and ref_view.view.space != reference.metadata.space
            ):
                continue
            new_entity = ViewEntity(
                # MyPy we validate that version is string in the constructor
                space=self.new_model_id.space,
                externalId=ref_view.view.external_id,
                version=self.new_model_id.version,  # type: ignore[arg-type]
            )
            if self.properties == "connection":
                # Set suffix to existing view and introduce a new view.
                # This will be used to point to the one view in the Enterprise model,
                # while the new view will to be written to.
                new_entity.suffix = f"{self.view_prefix}{ref_view.view.suffix}"
                new_view = DMSView(
                    view=ViewEntity(
                        # MyPy we validate that version is string in the constructor
                        space=self.new_model_id.space,
                        externalId=ref_view.view.external_id,
                        version=self.new_model_id.version,  # type: ignore[arg-type]
                    )
                )
                new_views.append(new_view)
                read_view_by_new_view[new_view.view] = new_entity
            elif self.properties == "repeat":
                # This is a slight misuse of the read_view_by_new_view. For the repeat mode, we only
                # care about the keys in this dictionary. But instead of creating a separate set, we
                # do this.
                read_view_by_new_view[new_entity] = new_entity

            renaming[ref_view.view] = new_entity
            new_views.append(ref_view.model_copy(deep=True, update={"implements": None, "view": new_entity}))

        new_properties = SheetList[DMSProperty]()
        new_view_entities = {view.view for view in new_views}
        for prop in reference.properties:
            new_property = prop.model_copy(deep=True)
            if new_property.value_type in renaming and isinstance(new_property.value_type, ViewEntity):
                new_property.value_type = renaming[new_property.value_type]
            if new_property.view in renaming:
                new_property.view = renaming[new_property.view]
            if new_property.view in new_view_entities and (
                not isinstance(new_property.value_type, ViewEntity) or new_property.value_type in new_view_entities
            ):
                new_properties.append(new_property)
        return new_views, new_properties, read_view_by_new_view

    def _create_containers_update_view_filter(
        self, new_views: SheetList[DMSView], reference: DMSRules, read_view_by_new_view: dict[ViewEntity, ViewEntity]
    ) -> tuple[SheetList[DMSContainer], SheetList[DMSProperty]]:
        new_containers = SheetList[DMSContainer]()
        container_properties: SheetList[DMSProperty] = SheetList[DMSProperty]()
        ref_containers_by_ref_view: dict[ViewEntity, set[ContainerEntity]] = defaultdict(set)
        ref_views_by_external_id = {
            view.view.external_id: view for view in reference.views if view.view.space == reference.metadata.space
        }
        if self.filter_type == "container":
            for prop in reference.properties:
                if prop.container:
                    ref_containers_by_ref_view[prop.view].add(prop.container)
        read_views = set(read_view_by_new_view.values())
        for view in new_views:
            if view.view in read_view_by_new_view:
                read_view = read_view_by_new_view[view.view]
                container_entity = ContainerEntity(space=self.new_model_id.space, externalId=view.view.external_id)
                prefix = to_camel_case(view.view.suffix)
                if self.properties == "repeat" and self.dummy_property:
                    property_ = DMSProperty(
                        view=view.view,
                        view_property=f"{prefix}{self.dummy_property}",
                        value_type=String(),
                        nullable=True,
                        immutable=False,
                        is_list=False,
                        container=container_entity,
                        container_property=f"{prefix}{self.dummy_property}",
                    )
                    new_containers.append(DMSContainer(container=container_entity))
                    container_properties.append(property_)
                elif self.properties == "repeat" and self.dummy_property is None:
                    # For this case we set the filter. This is used by the DataProductModel.
                    # Inherit view filter from original model to ensure the same instances are returned
                    # when querying the new view.
                    if ref_view := ref_views_by_external_id.get(view.view.external_id):
                        self._set_view_filter(view, ref_containers_by_ref_view, ref_view)
                elif self.properties == "connection" and self.direct_property:
                    property_ = DMSProperty(
                        view=view.view,
                        view_property=self.direct_property,
                        value_type=read_view,
                        nullable=True,
                        immutable=False,
                        is_list=False,
                        container=container_entity,
                        container_property=self.direct_property,
                    )
                    new_containers.append(DMSContainer(container=container_entity))
                    container_properties.append(property_)
                else:
                    raise NeatValueError(f"Unsupported properties mode: {self.properties}")

            if self.properties == "connection" and view.view in read_views:
                # Need to ensure that the 'Enterprise' view always returns the same instances
                # as the original view, no matter which properties are removed by the user.
                if ref_view := ref_views_by_external_id.get(view.view.external_id.removeprefix(self.view_prefix)):
                    self._set_view_filter(view, ref_containers_by_ref_view, ref_view)
        return new_containers, container_properties

    def _set_view_filter(
        self, view: DMSView, ref_containers_by_ref_view: dict[ViewEntity, set[ContainerEntity]], ref_view: DMSView
    ) -> None:
        if self.filter_type == "view":
            view.filter_ = HasDataFilter(inner=[ref_view.view])
        elif self.filter_type == "container" and (ref_containers := ref_containers_by_ref_view.get(ref_view.view)):
            # Sorting to ensure deterministic order
            view.filter_ = HasDataFilter(inner=sorted(ref_containers))


class ToDataProductModel(ToSolutionModel):
    """

    Args:
        new_model_id: DataData model identifier for the new model.
        include: The views to include in the data product data model. Can be either "same-space" or "all".
        filter_type: This is the type of filter to apply to the new views. The filter is used to
            ensure that the new views will return the same instance as the original views. The view filter is the
            simplest filter, but it has limitation in the fusion UI. The container filter is in essence a more
            verbose version of the view filter, and it has better support in the fusion UI. The default is "container".
    """

    type_: ClassVar[str] = "data_product"

    def __init__(
        self,
        new_model_id: DataModelIdentifier,
        include: Literal["same-space", "all"] = "same-space",
        filter_type: Literal["container", "view"] = "container",
        skip_cognite_views: bool = True,
    ):
        super().__init__(
            new_model_id,
            properties="repeat",
            dummy_property=None,
            filter_type=filter_type,
            exclude_views_in_other_spaces=include == "same-space",
            skip_cognite_views=skip_cognite_views,
        )
        self.include = include

    def transform(self, rules: DMSRules) -> DMSRules:
        # Overwrite transform to avoid the warning.
        return self._to_solution(rules)


class DropModelViews(VerifiedRulesTransformer[DMSRules, DMSRules]):
    _ASSET_VIEW = ViewId("cdf_cdm", "CogniteAsset", "v1")
    _VIEW_BY_COLLECTION: Mapping[Literal["3D", "Annotation", "BaseViews"], frozenset[ViewId]] = {
        "3D": frozenset(
            {
                ViewId("cdf_cdm", "Cognite3DModel", "v1"),
                ViewId("cdf_cdm", "Cognite3DObject", "v1"),
                ViewId("cdf_cdm", "Cognite3DRevision", "v1"),
                ViewId("cdf_cdm", "Cognite3DTransformation", "v1"),
                ViewId("cdf_cdm", "Cognite360Image", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageAnnotation", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageCollection", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageModel", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageStation", "v1"),
                ViewId("cdf_cdm", "CogniteCADModel", "v1"),
                ViewId("cdf_cdm", "CogniteCADNode", "v1"),
                ViewId("cdf_cdm", "CogniteCADRevision", "v1"),
                ViewId("cdf_cdm", "CogniteCubeMap", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudModel", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudRevision", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudVolume", "v1"),
            }
        ),
        "Annotation": frozenset(
            {
                ViewId("cdf_cdm", "CogniteAnnotation", "v1"),
                ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
            }
        ),
        "BaseViews": frozenset(
            {
                ViewId("cdf_cdm", "CogniteDescribable", "v1"),
                ViewId("cdf_cdm", "CogniteSchedulable", "v1"),
                ViewId("cdf_cdm", "CogniteSourceable", "v1"),
                ViewId("cdf_cdm", "CogniteVisualizable", "v1"),
            }
        ),
    }

    def __init__(
        self,
        view_external_id: str | SequenceNotStr[str] | None = None,
        group: Literal["3D", "Annotation", "BaseViews"]
        | Collection[Literal["3D", "Annotation", "BaseViews"]]
        | None = None,
    ):
        self.drop_external_ids = (
            {view_external_id} if isinstance(view_external_id, str) else set(view_external_id or [])
        )
        self.drop_collection = (
            [group]
            if isinstance(group, str)
            else cast(
                list[Literal["3D", "Annotation", "BaseViews"]],
                [collection for collection in group or [] if collection in self._VIEW_BY_COLLECTION],
            )
        )

    def transform(self, rules: DMSRules) -> DMSRules:
        exclude_views: set[ViewEntity] = {
            view.view for view in rules.views if view.view.suffix in self.drop_external_ids
        }
        if rules.metadata.as_data_model_id() in COGNITE_MODELS:
            exclude_views |= {
                ViewEntity.from_id(view_id, "v1")
                for collection in self.drop_collection
                for view_id in self._VIEW_BY_COLLECTION[collection]
            }
        new_model = rules.model_copy(deep=True)

        properties_by_view = RulesAnalysis(dms=new_model).properties_by_view(include_ancestors=True)

        new_model.views = SheetList[DMSView]([view for view in new_model.views if view.view not in exclude_views])
        new_properties = SheetList[DMSProperty]()
        mapped_containers: set[ContainerEntity] = set()
        for view in new_model.views:
            for prop in properties_by_view[view.view]:
                if "3D" in self.drop_collection and self._is_asset_3D_property(prop):
                    # We filter out the 3D property of asset
                    continue
                if isinstance(prop.value_type, ViewEntity) and prop.value_type in exclude_views:
                    continue
                new_properties.append(prop)
                if prop.container:
                    mapped_containers.add(prop.container)

        new_model.properties = new_properties
        new_model.containers = (
            SheetList[DMSContainer](
                [container for container in new_model.containers or [] if container.container in mapped_containers]
            )
            or None
        )

        return new_model

    @classmethod
    def _is_asset_3D_property(cls, prop: DMSProperty) -> bool:
        return prop.view.as_id() == cls._ASSET_VIEW and prop.view_property == "object3D"

    @property
    def description(self) -> str:
        return f"Removed {len(self.drop_external_ids) + len(self.drop_collection)} views from data model"


class IncludeReferenced(VerifiedRulesTransformer[DMSRules, DMSRules]):
    def __init__(self, client: NeatClient, include_properties: bool = False) -> None:
        self._client = client
        self.include_properties = include_properties

    def transform(self, rules: DMSRules) -> DMSRules:
        dms_rules = rules
        view_ids, container_ids = DMSValidation(dms_rules).imported_views_and_containers_ids()
        if not (view_ids or container_ids):
            warnings.warn(
                NeatValueWarning(
                    f"Data model {dms_rules.metadata.as_data_model_id()} does not have any "
                    "referenced views or containers."
                    "that is not already included in the data model."
                ),
                stacklevel=2,
            )
            return dms_rules

        schema = self._client.schema.retrieve([v.as_id() for v in view_ids], [c.as_id() for c in container_ids])
        copy_ = dms_rules.model_copy(deep=True)
        # Sorting to ensure deterministic order
        schema.containers = ContainerApplyDict(sorted(schema.containers.items(), key=lambda x: x[0].as_tuple()))
        schema.views = ViewApplyDict(sorted(schema.views.items(), key=lambda x: x[0].as_tuple()))
        importer = DMSImporter(schema)

        imported = importer.to_rules()
        if imported.rules is None:
            raise NeatValueError("Could not import the referenced views and containers.")

        verified = VerifyDMSRules(validate=False).transform(imported)
        if copy_.containers is None:
            copy_.containers = verified.containers
        else:
            existing_containers = {c.container for c in copy_.containers}
            copy_.containers.extend([c for c in verified.containers or [] if c.container not in existing_containers])
        existing_views = {v.view for v in copy_.views}
        copy_.views.extend([v for v in verified.views if v.view not in existing_views])
        if self.include_properties:
            existing_properties = {(p.view, p.view_property) for p in copy_.properties}
            copy_.properties.extend(
                [p for p in verified.properties if (p.view, p.view_property) not in existing_properties]
            )

        return copy_

    @property
    def description(self) -> str:
        return "Included referenced views and containers in the data model."


class AddClassImplements(VerifiedRulesTransformer[InformationRules, InformationRules]):
    def __init__(self, implements: str, suffix: str):
        self.implements = implements
        self.suffix = suffix

    def transform(self, rules: InformationRules) -> InformationRules:
        info_rules = rules
        output = info_rules.model_copy(deep=True)
        for class_ in output.classes:
            if class_.class_.suffix.endswith(self.suffix):
                class_.implements = [ClassEntity(prefix=class_.class_.prefix, suffix=self.implements)]
        return output

    @property
    def description(self) -> str:
        return f"Added implements property to classes with suffix {self.suffix}"


class ClassicPrepareCore(VerifiedRulesTransformer[InformationRules, InformationRules]):
    """Update the classic data model with the following:

    This is a special purpose transformer that is only intended to be used with when reading
    from classic cdf using the neat.read.cdf.classic.graph(...).

    - ClassicTimeseries.isString from boolean to string
    - Add class ClassicSourceSystem, and update all source properties from string to ClassicSourceSystem.
    - Rename externalId properties to classicExternalId
    - Renames the Relationship.sourceExternalId and Relationship.targetExternalId to startNode and endNode
    - If reference_timeseries is True, the classicExternalId property of the TimeSeries class will change type
      from string to timeseries.
    - If reference_files is True, the classicExternalId property of the File class will change type from string to file.
    """

    def __init__(
        self,
        instance_namespace: Namespace,
        reference_timeseries: bool = False,
        reference_files: bool = False,
    ) -> None:
        self.instance_namespace = instance_namespace
        self.reference_timeseries = reference_timeseries
        self.reference_files = reference_files

    @property
    def description(self) -> str:
        return "Update the classic data model to the data types in Cognite Core."

    def transform(self, rules: InformationRules) -> InformationRules:
        output = rules.model_copy(deep=True)
        for prop in output.properties:
            if prop.class_.suffix == "Timeseries" and prop.property_ == "isString":
                prop.value_type = String()
        prefix = output.metadata.prefix
        namespace = output.metadata.namespace
        source_system_class = InformationClass(
            class_=ClassEntity(prefix=prefix, suffix="ClassicSourceSystem"),
            description="A source system that provides data to the data model.",
            neatId=namespace["ClassicSourceSystem"],
            instance_source=self.instance_namespace["ClassicSourceSystem"],
        )
        output.classes.append(source_system_class)
        for prop in output.properties:
            if prop.property_ == "source" and prop.class_.suffix != "ClassicSourceSystem":
                prop.value_type = ClassEntity(prefix=prefix, suffix="ClassicSourceSystem")
            elif prop.property_ == "externalId":
                prop.property_ = "classicExternalId"
                if self.reference_timeseries and prop.class_.suffix == "ClassicTimeSeries":
                    prop.value_type = Timeseries()
                elif self.reference_files and prop.class_.suffix == "ClassicFile":
                    prop.value_type = File()
            elif prop.property_ == "sourceExternalId" and prop.class_.suffix == "ClassicRelationship":
                prop.property_ = "startNode"
            elif prop.property_ == "targetExternalId" and prop.class_.suffix == "ClassicRelationship":
                prop.property_ = "endNode"
        instance_prefix = next(
            (prefix for prefix, namespace in output.prefixes.items() if namespace == self.instance_namespace), None
        )
        if instance_prefix is None:
            raise NeatValueError("Instance namespace not found in the prefixes.")

        output.properties.append(
            InformationProperty(
                neatId=namespace["ClassicSourceSystem/name"],
                property_="name",
                value_type=String(),
                class_=ClassEntity(prefix=prefix, suffix="ClassicSourceSystem"),
                max_count=1,
                instance_source=[self.instance_namespace["name"]],
            )
        )
        return output


class ChangeViewPrefix(VerifiedRulesTransformer[DMSRules, DMSRules]):
    def __init__(self, old: str, new: str) -> None:
        self.old = old
        self.new = new

    def transform(self, rules: DMSRules) -> DMSRules:
        output = rules.model_copy(deep=True)
        new_by_old: dict[ViewEntity, ViewEntity] = {}
        for view in output.views:
            if view.view.external_id.startswith(self.old):
                new_external_id = f"{self.new}{view.view.external_id.removeprefix(self.old)}"
                new_view_entity = view.view.copy(update={"suffix": new_external_id})
                new_by_old[view.view] = new_view_entity
                view.view = new_view_entity
        for view in output.views:
            if view.implements:
                view.implements = [new_by_old.get(implemented, implemented) for implemented in view.implements]
        for prop in output.properties:
            if prop.view in new_by_old:
                prop.view = new_by_old[prop.view]
            if prop.value_type in new_by_old and isinstance(prop.value_type, ViewEntity):
                prop.value_type = new_by_old[prop.value_type]
        return output


class MergeDMSRules(VerifiedRulesTransformer[DMSRules, DMSRules]):
    def __init__(self, extra: DMSRules) -> None:
        self.extra = extra

    def transform(self, rules: DMSRules) -> DMSRules:
        output = rules.model_copy(deep=True)
        existing_views = {view.view for view in output.views}
        for view in self.extra.views:
            if view.view not in existing_views:
                output.views.append(view)
        existing_properties = {(prop.view, prop.view_property) for prop in output.properties}
        existing_containers = {container.container for container in output.containers or []}
        existing_enum_collections = {collection.collection for collection in output.enum or []}
        new_containers_by_entity = {container.container: container for container in self.extra.containers or []}
        new_enum_collections_by_entity = {collection.collection: collection for collection in self.extra.enum or []}
        for prop in self.extra.properties:
            if (prop.view, prop.view_property) in existing_properties:
                continue
            output.properties.append(prop)
            if prop.container and prop.container not in existing_containers:
                if output.containers is None:
                    output.containers = SheetList[DMSContainer]()
                output.containers.append(new_containers_by_entity[prop.container])
            if isinstance(prop.value_type, Enum) and prop.value_type.collection not in existing_enum_collections:
                if output.enum is None:
                    output.enum = SheetList[DMSEnum]()
                output.enum.append(new_enum_collections_by_entity[prop.value_type.collection])

        existing_nodes = {node.node for node in output.nodes or []}
        for node in self.extra.nodes or []:
            if node.node not in existing_nodes:
                if output.nodes is None:
                    output.nodes = SheetList[DMSNode]()
                output.nodes.append(node)

        return output

    @property
    def description(self) -> str:
        return f"Merged with {self.extra.metadata.as_data_model_id()}"


class MergeInformationRules(VerifiedRulesTransformer[InformationRules, InformationRules]):
    def __init__(self, extra: InformationRules) -> None:
        self.extra = extra

    def transform(self, rules: InformationRules) -> InformationRules:
        output = rules.model_copy(deep=True)
        existing_classes = {cls.class_ for cls in output.classes}
        for cls in self.extra.classes:
            if cls.class_ not in existing_classes:
                output.classes.append(cls)
        existing_properties = {(prop.class_, prop.property_) for prop in output.properties}
        for prop in self.extra.properties:
            if (prop.class_, prop.property_) not in existing_properties:
                output.properties.append(prop)
        for prefix, namespace in self.extra.prefixes.items():
            if prefix not in output.prefixes:
                output.prefixes[prefix] = namespace
        return output


class _InformationRulesConverter:
    _start_or_end_node: ClassVar[frozenset[str]] = frozenset({"endNode", "end_node", "startNode", "start_node"})

    def __init__(self, information: InformationRules):
        self.rules = information
        self.property_count_by_container: dict[ContainerEntity, int] = defaultdict(int)

    def as_dms_rules(
        self, ignore_undefined_value_types: bool = False, reserved_properties: Literal["error", "warning"] = "error"
    ) -> "DMSRules":
        from cognite.neat._rules.models.dms._rules import (
            DMSContainer,
            DMSProperty,
            DMSRules,
            DMSView,
        )

        info_metadata = self.rules.metadata
        default_version = info_metadata.version
        default_space = self._to_space(info_metadata.prefix)
        dms_metadata = self._convert_metadata_to_dms(info_metadata)

        properties_by_class: dict[ClassEntity, set[str]] = defaultdict(set)
        for prop in self.rules.properties:
            properties_by_class[prop.class_].add(prop.property_)

        # Edge Classes is defined by having both startNode and endNode properties
        edge_classes = {
            cls_
            for cls_, class_properties in properties_by_class.items()
            if ({"startNode", "start_node"} & class_properties) and ({"endNode", "end_node"} & class_properties)
        }
        edge_value_types_by_class_property_pair = {
            (prop.class_, prop.property_): prop.value_type
            for prop in self.rules.properties
            if prop.value_type in edge_classes and isinstance(prop.value_type, ClassEntity)
        }
        end_node_by_edge = {
            prop.class_: prop.value_type
            for prop in self.rules.properties
            if prop.class_ in edge_classes
            and (prop.property_ == "endNode" or prop.property_ == "end_node")
            and isinstance(prop.value_type, ClassEntity)
        }

        properties_by_class: dict[ClassEntity, list[DMSProperty]] = defaultdict(list)
        referenced_containers: dict[ContainerEntity, Counter[ClassEntity]] = defaultdict(Counter)
        for prop in self.rules.properties:
            if ignore_undefined_value_types and isinstance(prop.value_type, UnknownEntity):
                continue
            if prop.class_ in edge_classes and prop.property_ in self._start_or_end_node:
                continue
            if prop.property_ in DMS_RESERVED_PROPERTIES:
                msg = f"Property {prop.property_} is a reserved property in DMS."
                if reserved_properties == "error":
                    raise NeatValueError(msg)
                warnings.warn(NeatValueWarning(f"{msg} Skipping..."), stacklevel=2)
                continue

            dms_property = self._as_dms_property(
                prop,
                default_space,
                default_version,
                edge_classes,
                edge_value_types_by_class_property_pair,
                end_node_by_edge,
            )
            properties_by_class[prop.class_].append(dms_property)
            if dms_property.container:
                referenced_containers[dms_property.container][prop.class_] += 1

        views: list[DMSView] = []

        for cls_ in self.rules.classes:
            dms_view = DMSView(
                name=cls_.name,
                view=cls_.class_.as_view_entity(default_space, default_version),
                description=cls_.description,
                implements=self._get_view_implements(cls_, info_metadata),
            )

            dms_view.logical = cls_.neatId
            views.append(dms_view)

        class_by_entity = {cls_.class_: cls_ for cls_ in self.rules.classes}

        existing_containers: set[ContainerEntity] = set()

        containers: list[DMSContainer] = []
        for container_entity, class_entities in referenced_containers.items():
            if container_entity in existing_containers:
                continue
            constrains = self._create_container_constraint(
                class_entities, default_space, class_by_entity, referenced_containers
            )
            most_used_class_entity = class_entities.most_common(1)[0][0]
            class_ = class_by_entity[most_used_class_entity]

            if len(set(class_entities) - set(edge_classes)) == 0:
                used_for: Literal["node", "edge", "all"] = "edge"
            elif len(set(class_entities) - set(edge_classes)) == len(class_entities):
                used_for = "node"
            else:
                used_for = "all"

            container = DMSContainer(
                container=container_entity,
                name=class_.name,
                description=class_.description,
                constraint=constrains or None,
                used_for=used_for,
            )
            containers.append(container)

        dms_rules = DMSRules(
            metadata=dms_metadata,
            properties=SheetList[DMSProperty]([prop for prop_set in properties_by_class.values() for prop in prop_set]),
            views=SheetList[DMSView](views),
            containers=SheetList[DMSContainer](containers),
        )

        self.rules.sync_with_dms_rules(dms_rules)

        return dms_rules

    @staticmethod
    def _create_container_constraint(
        class_entities: Counter[ClassEntity],
        default_space: str,
        class_by_entity: dict[ClassEntity, InformationClass],
        referenced_containers: Collection[ContainerEntity],
    ) -> list[ContainerEntity]:
        constrains: list[ContainerEntity] = []
        for entity in class_entities:
            class_ = class_by_entity[entity]
            for parent in class_.implements or []:
                parent_entity = parent.as_container_entity(default_space)
                if parent_entity in referenced_containers:
                    constrains.append(parent_entity)
        return constrains

    @classmethod
    def _convert_metadata_to_dms(cls, metadata: InformationMetadata) -> "DMSMetadata":
        from cognite.neat._rules.models.dms._rules import (
            DMSMetadata,
        )

        dms_metadata = DMSMetadata(
            space=metadata.space,
            version=metadata.version,
            external_id=metadata.external_id,
            creator=metadata.creator,
            name=metadata.name,
            created=metadata.created,
            updated=metadata.updated,
        )

        dms_metadata.logical = metadata.identifier
        return dms_metadata

    def _as_dms_property(
        self,
        info_property: InformationProperty,
        default_space: str,
        default_version: str,
        edge_classes: set[ClassEntity],
        edge_value_types_by_class_property_pair: dict[tuple[ClassEntity, str], ClassEntity],
        end_node_by_edge: dict[ClassEntity, ClassEntity],
    ) -> "DMSProperty":
        from cognite.neat._rules.models.dms._rules import DMSProperty

        # returns property type, which can be ObjectProperty or DatatypeProperty
        value_type = self._get_value_type(
            info_property,
            default_space,
            default_version,
            edge_classes,
            end_node_by_edge,
        )

        connection = self._get_connection(
            info_property, value_type, edge_value_types_by_class_property_pair, default_space, default_version
        )

        container: ContainerEntity | None = None
        container_property: str | None = None
        is_list: bool | None = info_property.is_list
        nullable: bool | None = not info_property.is_mandatory
        if isinstance(connection, EdgeEntity):
            nullable = None
        elif connection == "direct":
            nullable = True
            container, container_property = self._get_container(info_property, default_space)
        else:
            container, container_property = self._get_container(info_property, default_space)

        dms_property = DMSProperty(
            name=info_property.name,
            value_type=value_type,
            nullable=nullable,
            is_list=is_list,
            connection=connection,
            default=info_property.default,
            container=container,
            container_property=container_property,
            view=info_property.class_.as_view_entity(default_space, default_version),
            view_property=info_property.property_,
        )

        # linking
        dms_property.logical = info_property.neatId

        return dms_property

    @staticmethod
    def _get_connection(
        prop: InformationProperty,
        value_type: DataType | ViewEntity | DMSUnknownEntity,
        edge_value_types_by_class_property_pair: dict[tuple[ClassEntity, str], ClassEntity],
        default_space: str,
        default_version: str,
    ) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None:
        if (
            isinstance(value_type, ViewEntity)
            and (prop.class_, prop.property_) in edge_value_types_by_class_property_pair
        ):
            edge_value_type = edge_value_types_by_class_property_pair[(prop.class_, prop.property_)]
            return EdgeEntity(properties=edge_value_type.as_view_entity(default_space, default_version))
        if isinstance(value_type, ViewEntity) and prop.is_list:
            return EdgeEntity()
        elif isinstance(value_type, ViewEntity):
            return "direct"
        # defaulting to direct connection
        elif isinstance(value_type, DMSUnknownEntity):
            return "direct"
        return None

    def _get_value_type(
        self,
        prop: InformationProperty,
        default_space: str,
        default_version: str,
        edge_classes: set[ClassEntity],
        end_node_by_edge: dict[ClassEntity, ClassEntity],
    ) -> DataType | ViewEntity | DMSUnknownEntity:
        if isinstance(prop.value_type, DataType):
            return prop.value_type

        # UnknownEntity should  resolve to DMSUnknownEntity
        # meaning end node type is unknown
        elif isinstance(prop.value_type, UnknownEntity):
            return DMSUnknownEntity()

        elif isinstance(prop.value_type, ClassEntity) and (prop.value_type in edge_classes):
            if prop.value_type in end_node_by_edge:
                return end_node_by_edge[prop.value_type].as_view_entity(default_space, default_version)
            # This occurs if the end node is not pointing to a class
            warnings.warn(
                NeatValueWarning(
                    f"Edge class {prop.value_type} does not have 'endNode' property, defaulting to DMSUnknownEntity"
                ),
                stacklevel=2,
            )
            return DMSUnknownEntity()
        elif isinstance(prop.value_type, ClassEntity):
            return prop.value_type.as_view_entity(default_space, default_version)

        elif isinstance(prop.value_type, MultiValueTypeInfo):
            # Multi Object type should resolve to DMSUnknownEntity
            # meaning end node type is unknown
            if prop.value_type.is_multi_object_type():
                non_unknown = [type_ for type_ in prop.value_type.types if isinstance(type_, UnknownEntity)]
                if list(non_unknown) == 1:
                    #
                    return non_unknown[0].as_view_entity(default_space, default_version)
                return DMSUnknownEntity()

            # Multi Data type should resolve to a single data type, or it should
            elif prop.value_type.is_multi_data_type():
                return self.convert_multi_data_type(prop.value_type)

            # Mixed types default to string
            else:
                non_any_uri = [type_ for type_ in prop.value_type.types if type_ != AnyURI()]
                if list(non_any_uri) == 1:
                    if isinstance(non_any_uri[0], ClassEntity):
                        return non_any_uri[0].as_view_entity(default_space, default_version)
                    else:
                        return non_any_uri[0]
                return String()

        raise ValueError(f"Unsupported value type: {prop.value_type.type_}")

    @classmethod
    def _to_space(cls, prefix: str) -> str:
        """Ensures that the prefix comply with the CDF space regex"""
        prefix = re.sub(r"[^a-zA-Z0-9_-]", "_", prefix)
        if prefix[0].isdigit() or prefix[0] == "_":
            prefix = f"a{prefix}"
        prefix = prefix[:43]
        if prefix[-1] == "_":
            prefix = f"{prefix[:-1]}1"
        return prefix

    def _get_container(self, prop: InformationProperty, default_space: str) -> tuple[ContainerEntity, str]:
        container_entity = prop.class_.as_container_entity(default_space)

        while self.property_count_by_container[container_entity] >= DMS_CONTAINER_PROPERTY_SIZE_LIMIT:
            container_entity.suffix = self._bump_suffix(container_entity.suffix)

        self.property_count_by_container[container_entity] += 1
        return container_entity, prop.property_

    def _get_view_implements(self, cls_: InformationClass, metadata: InformationMetadata) -> list[ViewEntity]:
        implements = []
        for parent in cls_.implements or []:
            view_entity = parent.as_view_entity(metadata.prefix, metadata.version)
            implements.append(view_entity)
        return implements

    @staticmethod
    def _bump_suffix(suffix: str) -> str:
        suffix_number = re.search(r"\d+$", suffix)

        if suffix_number:
            return suffix[: suffix_number.start()] + str(int(suffix_number.group()) + 1)
        else:
            return f"{suffix}2"

    @staticmethod
    def convert_multi_data_type(value_type: MultiValueTypeInfo) -> DataType:
        if not value_type.is_multi_data_type():
            raise ValueError("Only MultiValueType with DataType types is supported")
        # We check above that there are no ClassEntity types in the MultiValueType
        py_types = {type_.python for type_ in value_type.types}  # type: ignore[union-attr]

        # JSON mixed with other types should resolve to string that is safe choice
        if dms.Json in py_types and len(py_types) > 1:
            return data_types.String()
        elif dms.Json in py_types:
            return data_types.Json()
        elif not (py_types - {bool}):
            return data_types.Boolean()
        elif not (py_types - {int, bool}):
            return data_types.Integer()
        elif not (py_types - {float, int, bool}):
            return data_types.Double()
        elif not (py_types - {datetime, date}):
            return data_types.DateTime()

        return data_types.String()


class _DMSRulesConverter:
    def __init__(self, dms: DMSRules, instance_namespace: Namespace | None = None) -> None:
        self.dms = dms
        self.instance_namespace = instance_namespace

    def as_information_rules(
        self,
    ) -> "InformationRules":
        from cognite.neat._rules.models.information._rules import (
            InformationClass,
            InformationProperty,
            InformationRules,
        )

        dms = self.dms.metadata

        metadata = self._convert_metadata_to_info(dms)

        classes: list[InformationClass] = []
        for view in self.dms.views:
            info_class = InformationClass(
                # we do not want a version in class as we use URI for the class
                class_=ClassEntity(prefix=view.view.prefix, suffix=view.view.suffix),
                description=view.description,
                implements=[
                    # we do not want a version in class as we use URI for the class
                    implemented_view.as_class(skip_version=True)
                    for implemented_view in view.implements or []
                ],
            )

            # Linking
            info_class.physical = view.neatId
            classes.append(info_class)

        prefixes = get_default_prefixes_and_namespaces()
        if self.instance_namespace:
            instance_prefix = next((k for k, v in prefixes.items() if v == self.instance_namespace), None)
            if instance_prefix is None:
                # We need to add a new prefix
                instance_prefix = f"prefix_{len(prefixes) + 1}"
                prefixes[instance_prefix] = self.instance_namespace

        properties: list[InformationProperty] = []
        value_type: DataType | ClassEntity | str
        for property_ in self.dms.properties:
            if isinstance(property_.value_type, DataType):
                value_type = property_.value_type
            elif isinstance(property_.value_type, ViewEntity):
                value_type = ClassEntity(
                    prefix=property_.value_type.prefix,
                    suffix=property_.value_type.suffix,
                )
            elif isinstance(property_.value_type, DMSUnknownEntity):
                value_type = UnknownEntity()
            else:
                raise ValueError(f"Unsupported value type: {property_.value_type.type_}")

            info_property = InformationProperty(
                # Removing version
                class_=ClassEntity(suffix=property_.view.suffix, prefix=property_.view.prefix),
                property_=property_.view_property,
                value_type=value_type,
                description=property_.description,
                min_count=(0 if property_.nullable or property_.nullable is None else 1),
                max_count=(float("inf") if property_.is_list or property_.nullable is None else 1),
            )

            # Linking
            info_property.physical = property_.neatId

            properties.append(info_property)

        info_rules = InformationRules(
            metadata=metadata,
            properties=SheetList[InformationProperty](properties),
            classes=SheetList[InformationClass](classes),
            prefixes=prefixes,
        )

        self.dms.sync_with_info_rules(info_rules)

        return info_rules

    @classmethod
    def _convert_metadata_to_info(cls, metadata: DMSMetadata) -> "InformationMetadata":
        from cognite.neat._rules.models.information._rules import InformationMetadata

        return InformationMetadata(
            space=metadata.space,
            external_id=metadata.external_id,
            version=metadata.version,
            description=metadata.description,
            name=metadata.name,
            creator=metadata.creator,
            created=metadata.created,
            updated=metadata.updated,
        )


class SubsetDMSRules(VerifiedRulesTransformer[DMSRules, DMSRules]):
    """Subsets DMSRules to only include the specified views."""

    def __init__(self, views: set[ViewEntity]):
        self._views = views

    def transform(self, rules: DMSRules) -> DMSRules:
        analysis = RulesAnalysis(dms=rules)

        views_by_view = analysis.view_by_view_entity
        implements_by_view = analysis.implements_by_view()

        available = analysis.defined_views(include_ancestors=True)
        subset = available.intersection(self._views)

        ancestors: set[ViewEntity] = set()
        for view in subset:
            ancestors = ancestors.union({ancestor for ancestor in get_inheritance_path(view, implements_by_view)})
        subset = subset.union(ancestors)

        if not subset:
            raise NeatValueError("None of the requested views are defined in the rules!")

        if nonexisting := self._views - subset:
            raise NeatValueError(
                "Following requested views do not exist"
                f" in the rules: [{','.join([view.external_id for view in nonexisting])}]. Aborting."
            )

        subsetted_rules: dict[str, Any] = {
            "metadata": rules.metadata.model_copy(),
            "views": SheetList[DMSView](),
            "properties": SheetList[DMSProperty](),
            "containers": SheetList[DMSContainer](),
            "enum": rules.enum,
            "nodes": rules.nodes,
        }

        # add views
        for view in subset:
            subsetted_rules["views"].append(views_by_view[view])

        used_containers = set()

        # add properties
        for view, properties in analysis.properties_by_view(include_ancestors=False).items():
            if view not in subset:
                continue

            for property_ in properties:
                if (
                    isinstance(property_.value_type, DataType)
                    or isinstance(property_.value_type, DMSUnknownEntity)
                    or (isinstance(property_.value_type, ViewEntity) and property_.value_type in subset)
                ):
                    subsetted_rules["properties"].append(property_)

                    if property_.container:
                        used_containers.add(property_.container)

        # add containers
        if rules.containers:
            for container in rules.containers:
                if container.container in used_containers:
                    subsetted_rules["containers"].append(container)

        try:
            return DMSRules.model_validate(subsetted_rules)
        except ValidationError as e:
            raise NeatValueError(f"Cannot subset rules: {e}") from e


class SubsetInformationRules(VerifiedRulesTransformer[InformationRules, InformationRules]):
    """Subsets InformationRules to only include the specified classes."""

    def __init__(self, classes: set[ClassEntity]):
        self._classes = classes

    def transform(self, rules: InformationRules) -> InformationRules:
        analysis = RulesAnalysis(information=rules)

        class_by_class_entity = analysis.class_by_class_entity
        parent_entity_by_class_entity = analysis.parents_by_class()

        available = analysis.defined_classes(include_ancestors=True)
        subset = available.intersection(self._classes)

        # need to add all the parent classes of the desired classes to the possible classes
        ancestors: set[ClassEntity] = set()
        for class_ in subset:
            ancestors = ancestors.union(
                {ancestor for ancestor in get_inheritance_path(class_, parent_entity_by_class_entity)}
            )
        subset = subset.union(ancestors)

        if not subset:
            raise NeatValueError("None of the requested classes are defined in the rules!")

        if nonexisting := self._classes - subset:
            raise NeatValueError(
                "Following requested classes do not exist"
                f" in the rules: [{','.join([class_.suffix for class_ in nonexisting])}]"
                ". Aborting."
            )

        subsetted_rules: dict[str, Any] = {
            "metadata": rules.metadata.model_copy(),
            "prefixes": (rules.prefixes or {}).copy(),
            "classes": SheetList[InformationClass](),
            "properties": SheetList[InformationProperty](),
        }

        for class_ in subset:
            subsetted_rules["classes"].append(class_by_class_entity[class_])

        for class_, properties in analysis.properties_by_class(include_ancestors=False).items():
            if class_ not in subset:
                continue
            for property_ in properties:
                # datatype property can be added directly
                if (
                    isinstance(property_.value_type, DataType)
                    or (isinstance(property_.value_type, ClassEntity) and property_.value_type in subset)
                    or isinstance(property_.value_type, UnknownEntity)
                ):
                    subsetted_rules["properties"].append(property_)
                # object property can be added if the value type is in the subset
                elif isinstance(property_.value_type, MultiValueTypeInfo):
                    allowed = [t for t in property_.value_type.types if t in subset or isinstance(t, DataType)]
                    if allowed:
                        subsetted_rules["properties"].append(
                            property_.model_copy(
                                deep=True,
                                update={"value_type": MultiValueTypeInfo(types=allowed)},
                            )
                        )

        try:
            return InformationRules.model_validate(subsetted_rules)
        except ValidationError as e:
            raise NeatValueError(f"Cannot subset rules: {e}") from e



"""Neat sequence combines the CogniteClient sequence and the SequenceRows"""

import sys
import warnings
from abc import ABC
from collections.abc import Sequence
from typing import Any, cast

import cognite.client.data_classes as cdc
from cognite.client import CogniteClient
from cognite.client.data_classes import (
    SequenceColumn,
    SequenceColumnList,
    SequenceColumnWrite,
    SequenceColumnWriteList,
    SequenceRow,
)
from cognite.client.data_classes._base import (
    CogniteResourceList,
    ExternalIDTransformerMixin,
    IdTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class NeatSequenceCore(WriteableCogniteResource["NeatSequenceWrite"], ABC):
    """Information about the sequence stored in the database

    Args:
        name (str | None): Name of the sequence
        description (str | None): Description of the sequence
        asset_id (int | None): Optional asset this sequence is associated with
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        metadata (dict[str, Any] | None): Custom, application-specific metadata. String key -> String value.
            The maximum length of the key is 32 bytes, the value 512 bytes, with up to 16 key-value pairs.
        data_set_id (int | None): Data set that this sequence belongs to
        rows (typing.Sequence[SequenceRow] | None): The rows in the sequence.
    """

    def __init__(
        self,
        name: str | None = None,
        description: str | None = None,
        asset_id: int | None = None,
        external_id: str | None = None,
        metadata: dict[str, Any] | None = None,
        data_set_id: int | None = None,
        rows: Sequence[SequenceRow] | None = None,
    ) -> None:
        self.name = name
        self.description = description
        self.asset_id = asset_id
        self.external_id = external_id
        self.metadata = metadata
        self.data_set_id = data_set_id
        self.rows = rows


class NeatSequenceWrite(NeatSequenceCore):
    """Information about the sequence stored in the database.
    This is the writing version of the class, it is used for inserting data into the CDF.

    Args:
        columns (typing.Sequence[SequenceColumnWrite]): List of column definitions
        name (str | None): Name of the sequence
        description (str | None): Description of the sequence
        asset_id (int | None): Optional asset this sequence is associated with
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        metadata (dict[str, Any] | None): Custom, application-specific metadata. String key -> String value.
            The maximum length of key is 32 bytes, value 512 bytes, up to 16 key-value pairs.
        data_set_id (int | None): Data set that this sequence belongs to
        rows (typing.Sequence[SequenceRow] | None): The rows in the sequence.
    """

    def __init__(
        self,
        columns: Sequence[SequenceColumnWrite],
        name: str | None = None,
        description: str | None = None,
        asset_id: int | None = None,
        external_id: str | None = None,
        metadata: dict[str, Any] | None = None,
        data_set_id: int | None = None,
        rows: Sequence[SequenceRow] | None = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            asset_id=asset_id,
            external_id=external_id,
            metadata=metadata,
            data_set_id=data_set_id,
            rows=rows,
        )
        self.columns: SequenceColumnWriteList
        if isinstance(columns, SequenceColumnWriteList):
            self.columns = columns
        elif isinstance(columns, Sequence) and all(isinstance(col, SequenceColumnWrite) for col in columns):
            self.columns = SequenceColumnWriteList(columns)
        else:
            raise ValueError(f"columns must be a sequence of SequenceColumnWrite objects not {type(columns)}")

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            columns=SequenceColumnWriteList._load(resource["columns"]),
            name=resource.get("name"),
            description=resource.get("description"),
            asset_id=resource.get("assetId"),
            external_id=resource.get("externalId"),
            metadata=resource.get("metadata"),
            data_set_id=resource.get("dataSetId"),
            rows=[SequenceRow._load(row) for row in resource.get("rows", [])],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        dumped = super().dump(camel_case)
        dumped["columns"] = self.columns.dump(camel_case)
        if self.rows is not None:
            dumped["rows"] = [row.dump(camel_case) for row in self.rows]
        return dumped

    def as_write(self) -> "NeatSequenceWrite":
        """Returns this NeatSequenceWrite."""
        return self


class NeatSequence(NeatSequenceCore):
    """Information about the sequence stored in the database.
    This is the reading version of the class, it is used for retrieving data from the CDF.

    Args:
        id (int): Unique cognite-provided identifier for the sequence
        name (str | None): Name of the sequence
        created_time (int | None): Time when this sequence was created in CDF in milliseconds since Jan 1, 1970.
        last_updated_time (int | None): The last time this sequence was updated in CDF,
            in milliseconds since Jan 1, 1970.
        description (str | None): Description of the sequence
        asset_id (int | None): Optional asset this sequence is associated with
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        metadata (dict[str, Any] | None): Custom, application-specific metadata. String key -> String value.
            The maximum length of the key is 32 bytes, the value 512 bytes, with up to 16 key-value pairs.
        columns (Sequence[SequenceColumn] | None): List of column definitions

        data_set_id (int | None): Data set that this sequence belongs to
        rows (Sequence[SequenceRow] | None): The rows in the sequence.
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        id: int,
        created_time: int,
        last_updated_time: int,
        name: str | None = None,
        description: str | None = None,
        asset_id: int | None = None,
        external_id: str | None = None,
        metadata: dict[str, Any] | None = None,
        columns: Sequence[SequenceColumn] | None = None,
        data_set_id: int | None = None,
        rows: Sequence[SequenceRow] | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(
            name=name,
            description=description,
            asset_id=asset_id,
            external_id=external_id,
            metadata=metadata,
            data_set_id=data_set_id,
            rows=rows,
        )
        self.id = id
        self.created_time = created_time
        self.last_updated_time = last_updated_time

        self.columns: SequenceColumnList | None
        if columns is None or isinstance(columns, SequenceColumnList):
            self.columns = columns
        elif isinstance(columns, Sequence) and all(isinstance(col, SequenceColumn) for col in columns):
            self.columns = SequenceColumnList(columns)
        elif isinstance(columns, list):
            warnings.warn(
                "Columns is no longer a dict, you should first load the list of dictionaries using "
                "SequenceColumnList.load([{...}, {...}])",
                DeprecationWarning,
                stacklevel=2,
            )
            self.columns = SequenceColumnList._load(columns)
        else:
            raise ValueError(f"columns must be a sequence of SequenceColumn objects not {type(columns)}")
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def from_cognite_sequence(
        cls, sequence: cdc.Sequence, rows: Sequence[cdc.SequenceRow] | None = None
    ) -> "NeatSequence":
        """Create a NeatSequence from a Cognite Sequence object."""
        args = sequence.dump(camel_case=True)
        if rows is not None:
            args["rows"] = [row.dump(camel_case=True) for row in rows]
        return cls._load(args)

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            name=resource.get("name"),
            description=resource.get("description"),
            asset_id=resource.get("assetId"),
            external_id=resource.get("externalId"),
            metadata=resource.get("metadata"),
            columns=SequenceColumnList._load(resource["columns"]) if "columns" in resource else None,
            data_set_id=resource.get("dataSetId"),
            rows=[SequenceRow._load(row) for row in resource.get("rows", [])] or None,
        )

    def as_write(self) -> "NeatSequenceWrite":
        """Returns a writeable version of this sequence."""
        if self.columns is None:
            raise ValueError("Columns must be set for the writing version of the sequence")

        return NeatSequenceWrite(
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            asset_id=self.asset_id,
            metadata=self.metadata,
            data_set_id=self.data_set_id,
            columns=self.columns.as_write(),
            rows=self.rows,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        dumped = super().dump(camel_case)
        if self.columns is not None:
            dumped["columns"] = self.columns.dump(camel_case)
        if self.rows is not None:
            dumped["rows"] = [row.dump(camel_case) for row in self.rows]
        return dumped


class NeatSequenceWriteList(CogniteResourceList[NeatSequenceWrite], ExternalIDTransformerMixin):
    _RESOURCE = NeatSequenceWrite


class NeatSequenceList(WriteableCogniteResourceList[NeatSequenceWrite, NeatSequence], IdTransformerMixin):
    _RESOURCE = NeatSequence

    def as_write(self) -> NeatSequenceWriteList:
        """Returns a writeable version of this sequence list."""
        return NeatSequenceWriteList([item.as_write() for item in self], cognite_client=self._get_cognite_client())



import difflib
from collections.abc import Collection
from typing import Literal

from cognite.client.utils.useful_types import SequenceNotStr
from rdflib import URIRef

from cognite.neat._constants import COGNITE_MODELS
from cognite.neat._issues import IssueList
from cognite.neat._rules.transformers import DropModelViews

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper

try:
    from rich import print
except ImportError:
    ...


@session_class_wrapper
class DropAPI:
    """
    Drop instances from the session. Check out `.instances()` for performing the operation.
    """

    def __init__(self, state: SessionState):
        self._state = state
        self.data_model = DropDataModelAPI(state)

    def instances(self, type: str | list[str]) -> None:
        """Drop instances from the session.

        Args:
            type: The type of instances to drop.

        Example:
            ```python
            node_type_to_drop = "Pump"
            neat.drop.instances(node_type_to_drop)
            ```
        """
        type_list = type if isinstance(type, list) else [type]

        # Temporary solution until we agree on the form of specifying named graphs
        # it will default to the default named graph
        named_graph = self._state.instances.store.default_named_graph
        uri_type_type = dict((v, k) for k, v in self._state.instances.store.queries.types(named_graph).items())
        selected_uri_by_type: dict[URIRef, str] = {}
        for type_item in type_list:
            if type_item not in uri_type_type:
                print(f"Type {type_item} not found.")
            selected_uri_by_type[uri_type_type[type_item]] = type_item

        result = self._state.instances.store.queries.drop_types(list(selected_uri_by_type.keys()))

        for type_uri, count in result.items():
            print(f"Dropped {count} instances of type {selected_uri_by_type[type_uri]}")
        return None


@session_class_wrapper
class DropDataModelAPI:
    def __init__(self, state: SessionState):
        self._state = state

    def views(
        self,
        view_external_id: str | SequenceNotStr[str] | None = None,
        group: Literal["3D", "Annotation", "BaseViews"]
        | Collection[Literal["3D", "Annotation", "BaseViews"]]
        | None = None,
    ) -> IssueList:
        """Drops views from the data model.
        Args:
            view_external_id: The externalId of the view to drop.
            group: Only applies to CogniteCore model. This is a shorthand for dropping multiple views at once.
        """
        if sum([view_external_id is not None, group is not None]) != 1:
            raise NeatSessionError("Only one of view_external_id or group can be specified.")
        last_dms = self._state.rule_store.last_verified_dms_rules
        if group is not None and last_dms.metadata.as_data_model_id() not in COGNITE_MODELS:
            raise NeatSessionError("Group can only be specified for CogniteCore models.")
        if view_external_id is not None:
            existing_views = {view.view.external_id for view in last_dms.views}
            requested_views = {view_external_id} if isinstance(view_external_id, str) else set(view_external_id)
            missing_views = requested_views - existing_views
            if missing_views:
                suggestions: list[str] = []
                for view in missing_views:
                    suggestion = difflib.get_close_matches(view, existing_views, n=1)
                    if suggestion:
                        suggestions.append(f"{view} -> {suggestion[0]}")
                    else:
                        suggestions.append(f"{view} -> NOT FOUND")
                raise NeatSessionError(
                    f"{len(missing_views)} view(s) not found in the data model.\nDid you mean {', '.join(suggestions)}?"
                )
        before = len(last_dms.views)
        issues = self._state.rule_transform(DropModelViews(view_external_id, group))
        after = len(self._state.rule_store.last_verified_dms_rules.views)
        print(f"Dropped {before - after} views.")
        return issues



from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import lru_cache
from pathlib import Path
from types import UnionType
from typing import TYPE_CHECKING, Generic, TypeVar, Union, get_args, get_origin

from cognite.neat._client import NeatClient
from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules._shared import T_VerifiedRules
from cognite.neat._utils.auxiliary import class_html_doc
from cognite.neat._utils.upload import UploadResult, UploadResultList

if TYPE_CHECKING:
    from cognite.neat._store._provenance import Agent as ProvenanceAgent

T_Export = TypeVar("T_Export")


class BaseExporter(ABC, Generic[T_VerifiedRules, T_Export]):
    _new_line = "\n"
    _encoding = "utf-8"

    @abstractmethod
    def export_to_file(self, rules: T_VerifiedRules, filepath: Path) -> None:
        raise NotImplementedError

    @abstractmethod
    def export(self, rules: T_VerifiedRules) -> T_Export:
        raise NotImplementedError

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls, include_factory_methods=False)

    @property
    def agent(self) -> "ProvenanceAgent":
        """Provenance agent for the importer."""
        from cognite.neat._store._provenance import Agent as ProvenanceAgent

        return ProvenanceAgent(id_=DEFAULT_NAMESPACE[f"agent/{type(self).__name__}"])

    @property
    def description(self) -> str:
        return "MISSING DESCRIPTION"

    @classmethod
    @lru_cache(maxsize=1)
    def source_types(cls) -> tuple[type, ...]:
        base_exporter = cls.__orig_bases__[0]  # type: ignore[attr-defined]
        source_type = get_args(base_exporter)[0]
        if get_origin(source_type) in [Union, UnionType]:
            return get_args(source_type)
        return (source_type,)


class CDFExporter(BaseExporter[T_VerifiedRules, T_Export], ABC):
    @abstractmethod
    def export_to_cdf_iterable(
        self, rules: T_VerifiedRules, client: NeatClient, dry_run: bool = False
    ) -> Iterable[UploadResult]:
        raise NotImplementedError

    def export_to_cdf(self, rules: T_VerifiedRules, client: NeatClient, dry_run: bool = False) -> UploadResultList:
        return UploadResultList(self.export_to_cdf_iterable(rules, client, dry_run))



from typing import cast

from rdflib import BNode, Graph
from rdflib.plugins.sparql import prepareQuery
from rdflib.query import ResultRow

from cognite.neat._issues._base import IssueList
from cognite.neat._issues.errors._general import NeatValueError
from cognite.neat._issues.warnings._resources import (
    ResourceRedefinedWarning,
    ResourceRetrievalWarning,
)
from cognite.neat._utils.rdf_ import convert_rdflib_content


def parse_classes(graph: Graph, query: str, language: str, issue_list: IssueList) -> tuple[dict, IssueList]:
    """Parse classes from graph

    Args:
        graph: Graph containing classes definitions
        language: Language to use for parsing, by default "en"

    Returns:
        Dataframe containing owl classes
    """

    classes: dict[str, dict] = {}

    query = prepareQuery(query.format(language=language), initNs={k: v for k, v in graph.namespaces()})
    expected_keys = [str(v) for v in query.algebra._vars]

    for raw in graph.query(query):
        res: dict = convert_rdflib_content(cast(ResultRow, raw).asdict(), True)
        res = {key: res.get(key, None) for key in expected_keys}

        class_id = res["class_"]

        # Safeguarding against incomplete semantic definitions
        if res["implements"] and isinstance(res["implements"], BNode):
            issue_list.append(
                ResourceRetrievalWarning(
                    class_id,
                    "implements",
                    error=("Unable to determine class that is being implemented"),
                )
            )
            continue

        if class_id not in classes:
            classes[class_id] = res
        else:
            # Handling implements
            if classes[class_id]["implements"] and isinstance(classes[class_id]["implements"], list):
                if res["implements"] not in classes[class_id]["implements"]:
                    classes[class_id]["implements"].append(res["implements"])

            elif classes[class_id]["implements"] and isinstance(classes[class_id]["implements"], str):
                classes[class_id]["implements"] = [classes[class_id]["implements"]]

                if res["implements"] not in classes[class_id]["implements"]:
                    classes[class_id]["implements"].append(res["implements"])
            elif res["implements"]:
                classes[class_id]["implements"] = [res["implements"]]

            handle_meta("class_", classes, class_id, res, "name", issue_list)
            handle_meta("class_", classes, class_id, res, "description", issue_list)

    if not classes:
        issue_list.append(NeatValueError("Unable to parse classes"))

    return classes, issue_list


def parse_properties(graph: Graph, query: str, language: str, issue_list: IssueList) -> tuple[dict, IssueList]:
    """Parse properties from graph

    Args:
        graph: Graph containing owl classes
        language: Language to use for parsing, by default "en"

    Returns:
        Dataframe containing owl classes
    """

    properties: dict[str, dict] = {}

    query = prepareQuery(query.format(language=language), initNs={k: v for k, v in graph.namespaces()})
    expected_keys = [str(v) for v in query.algebra._vars]

    for raw in graph.query(query):
        res: dict = convert_rdflib_content(cast(ResultRow, raw).asdict(), True)
        res = {key: res.get(key, None) for key in expected_keys}

        property_id = res["property_"]

        # Safeguarding against incomplete semantic definitions
        if not res["class_"] or isinstance(res["class_"], BNode):
            issue_list.append(
                ResourceRetrievalWarning(
                    property_id,
                    "property",
                    error=("Unable to determine to what class property is being defined"),
                )
            )
            continue

        # Safeguarding against incomplete semantic definitions
        if not res["value_type"] or isinstance(res["value_type"], BNode):
            issue_list.append(
                ResourceRetrievalWarning(
                    property_id,
                    "property",
                    error=("Unable to determine value type of property"),
                )
            )
            continue

        id_ = f"{res['class_']}.{res['property_']}"

        if id_ not in properties:
            properties[id_] = res
            properties[id_]["value_type"] = [properties[id_]["value_type"]]
        else:
            handle_meta("property", properties, id_, res, "name", issue_list)
            handle_meta(
                "property",
                properties,
                id_,
                res,
                "description",
                issue_list,
            )

            # Handling multi-value types
            if res["value_type"] not in properties[id_]["value_type"]:
                properties[id_]["value_type"].append(res["value_type"])

    for prop in properties.values():
        prop["value_type"] = ", ".join(prop["value_type"])

    if not properties:
        issue_list.append(NeatValueError("Unable to parse properties"))

    return properties, issue_list


def handle_meta(
    resource_type: str,
    resources: dict[str, dict],
    resource_id: str,
    res: dict,
    feature: str,
    issue_list: IssueList,
):
    if not resources[resource_id][feature] and res[feature]:
        resources[resource_id][feature] = res[feature]

    # RAISE warning only if the feature is being redefined
    elif resources[resource_id][feature] and res[feature]:
        issue_list.append(
            ResourceRedefinedWarning(
                identifier=resource_id,
                resource_type=resource_type,
                feature=feature,
                current_value=resources[resource_id][feature],
                new_value=res[feature],
            )
        )



from cognite.neat._client.data_classes.schema import DMSSchema
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._rules.models.information._rules_input import InformationInputRules

from ._base_rules import DataModelType, ExtensionCategory, RoleTypes, SchemaCompleteness, SheetList, SheetRow
from .dms._rules import DMSRules
from .dms._rules_input import DMSInputRules

INPUT_RULES_BY_ROLE: dict[RoleTypes, type[InformationInputRules] | type[DMSInputRules]] = {
    RoleTypes.information: InformationInputRules,
    RoleTypes.dms: DMSInputRules,
}
VERIFIED_RULES_BY_ROLE: dict[RoleTypes, type[InformationRules] | type[DMSRules]] = {
    RoleTypes.information: InformationRules,
    RoleTypes.dms: DMSRules,
}


__all__ = [
    "INPUT_RULES_BY_ROLE",
    "DMSInputRules",
    "DMSRules",
    "DMSSchema",
    "DataModelType",
    "ExtensionCategory",
    "InformationInputRules",
    "InformationRules",
    "RoleTypes",
    "SchemaCompleteness",
    "SheetList",
    "SheetRow",
]



"""These are special exceptions that are used by the store to signal invalid transformers"""

from dataclasses import dataclass

from cognite.neat._graph.extractors import KnowledgeGraphExtractor
from cognite.neat._issues import IssueList
from cognite.neat._rules.importers import BaseImporter
from cognite.neat._rules.transformers import VerifiedRulesTransformer

from ._provenance import Activity


class NeatStoreError(Exception):
    """Base class for all exceptions in the store module"""

    def __str__(self):
        return type(self).__name__


class ActivityFailed(NeatStoreError):
    """Raised when an activity fails"""

    def __init__(
        self,
        activity: Activity,
        issue_list: IssueList,
        tool: BaseImporter | VerifiedRulesTransformer | KnowledgeGraphExtractor,
    ) -> None:
        self.activity = activity
        self.issue_list = issue_list
        self.tool = tool

    def __str__(self):
        return self.tool.description


@dataclass
class InvalidActivityInput(NeatStoreError, RuntimeError):
    """Raised when an invalid activity is attempted"""

    expected: tuple[type, ...]
    have: tuple[type, ...]


class InvalidActivityOutput(NeatStoreError):
    """Raised when an activity has an invalid output"""

    def __init__(self, activity: Activity, output: str) -> None:
        self.activity = activity
        self.output = output

    def __str__(self):
        return f"{super().__str__()}: {self.activity.id_} -> {self.output}"


class EmptyStore(NeatStoreError, RuntimeError):
    """Raised when the store is empty"""

    ...



"""This module contains the definition of `TransformationRules` pydantic model and all
its sub-models and validators.
"""

import math
import sys
import types
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterator, MutableSequence, Sequence
from datetime import datetime
from typing import (
    Annotated,
    Any,
    ClassVar,
    Literal,
    SupportsIndex,
    TypeVar,
    get_args,
    get_origin,
    overload,
)

import pandas as pd
from cognite.client import data_modeling as dm
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    PlainSerializer,
    field_validator,
    model_serializer,
)
from pydantic.main import IncEx
from pydantic_core import core_schema
from rdflib import Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules.models._types import (
    ContainerEntityType,
    DataModelExternalIdType,
    DmsPropertyType,
    SpaceType,
    StrListType,
    URIRefType,
    VersionType,
    ViewEntityType,
)
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import EdgeEntity, ReverseConnectionEntity, ViewEntity
from cognite.neat._utils.rdf_ import uri_display_name

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


METADATA_VALUE_MAX_LENGTH = 5120


def _get_required_fields(model: type[BaseModel], use_alias: bool = False) -> set[str]:
    """Get required fields from a pydantic model.

    Parameters
    ----------
    model : type[BaseModel]
        Pydantic data model
    use_alias : bool, optional
        Whether to return field alias name, by default False

    Returns
    -------
    list[str]
        List of required fields
    """
    required_fields = set()
    for name, field in model.model_fields.items():
        if not field.is_required():
            continue

        alias = getattr(field, "alias", None)
        if use_alias and alias:
            required_fields.add(alias)
        else:
            required_fields.add(name)
    return required_fields


class SchemaCompleteness(StrEnum):
    complete = "complete"
    partial = "partial"
    extended = "extended"


class ExtensionCategory(StrEnum):
    addition = "addition"
    reshape = "reshape"
    rebuild = "rebuild"


class DataModelType(StrEnum):
    solution = "solution"
    enterprise = "enterprise"


class DataModelAspect(StrEnum):
    conceptual = "conceptual"
    logical = "logical"
    physical = "physical"


class RoleTypes(StrEnum):
    domain_expert = "domain expert"
    information = "information architect"
    dms = "DMS Architect"


class SchemaModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True,
        str_strip_whitespace=True,
        arbitrary_types_allowed=True,
        strict=False,
        extra="ignore",
        use_enum_values=True,
    )

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        elif isinstance(value, list):
            return [entry.strip() if isinstance(entry, str) else entry for entry in value]
        return value


class BaseMetadata(SchemaModel):
    """
    Metadata model for data model
    """

    role: ClassVar[RoleTypes] = Field(description="Role of the person creating the data model")
    aspect: ClassVar[DataModelAspect] = Field(description="Aspect of the data model")
    space: SpaceType = Field(description="The space where the data model is defined")
    external_id: DataModelExternalIdType = Field(
        alias="externalId", description="External identifier for the data model"
    )
    version: VersionType = Field(description="Version of the data model")

    name: str | None = Field(
        None,
        description="Human readable name of the data model",
        min_length=1,
        max_length=255,
    )

    description: str | None = Field(
        None, min_length=1, max_length=1024, description="Short description of the data model"
    )

    creator: StrListType = Field(
        description=(
            "List of contributors (comma seperated) to the data model creation, "
            "typically information architects are considered as contributors."
        ),
    )

    created: datetime = Field(
        description="Date of the data model creation",
    )

    updated: datetime = Field(
        description="Date of the data model update",
    )

    source_id: URIRefType | None = Field(
        None,
        description="Id of source that produced this rules",
        alias="sourceId",
    )

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        return value

    @field_validator("description", mode="before")
    def nan_as_none(cls, value):
        if isinstance(value, float) and math.isnan(value):
            return None
        return value

    def to_pandas(self) -> pd.Series:
        """Converts Metadata to pandas Series."""
        return pd.Series(self.model_dump())

    def _repr_html_(self) -> str:
        """Returns HTML representation of Metadata."""
        return self.to_pandas().to_frame("value")._repr_html_()  # type: ignore[operator]

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @model_serializer(mode="wrap")
    def include_role(self, serializer: Callable) -> dict:
        return {"role": self.role.value, **serializer(self)}

    @property
    def prefix(self) -> str:
        return self.space

    def get_prefix(self) -> str:
        return self.prefix

    @property
    def identifier(self) -> URIRef:
        """Globally unique identifier for the data model.

        !!! note
            Unlike namespace, the identifier does not end with "/" or "#".

        """
        return DEFAULT_NAMESPACE[f"data-model/verified/{self.aspect}/{self.space}/{self.external_id}/{self.version}"]

    @property
    def namespace(self) -> Namespace:
        """Namespace for the data model used for the entities in the data model."""
        return Namespace(f"{self.identifier}/")

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    def as_identifier(self) -> str:
        return repr(self.as_data_model_id())


class BaseRules(SchemaModel, ABC):
    """
    Rules is a core concept in `neat`. This represents fusion of data model
    definitions and (optionally) the transformation rules used to transform the data/graph
    from the source representation to the target representation defined by the data model.
    The rules are defined in an Excel sheet and then parsed into a `Rules` object. The
    `Rules` object is then used to generate data model and the `RDF` graph made of data
    model instances.

    Args:
        metadata: Data model metadata
    """

    metadata: BaseMetadata

    @classmethod
    def headers_by_sheet(cls, by_alias: bool = False) -> dict[str, list[str]]:
        """Returns a list of headers for the model, typically used by ExcelExporter"""
        headers_by_sheet: dict[str, list[str]] = {}
        for field_name, field in cls.model_fields.items():
            if field_name in ["validators_to_skip", "post_validate"]:
                continue
            sheet_name = (field.alias or field_name) if by_alias else field_name
            annotation = field.annotation

            if isinstance(annotation, types.UnionType):
                annotation = annotation.__args__[0]

            try:
                if isinstance(annotation, types.GenericAlias) and get_origin(annotation) is SheetList:
                    # We know that this is a SheetList, so we can safely access the annotation
                    # which is the concrete type of the SheetEntity.
                    model_fields = get_args(annotation)[0].model_fields  # type: ignore[union-attr]
                elif isinstance(annotation, type) and issubclass(annotation, BaseModel):
                    model_fields = annotation.model_fields
                else:
                    model_fields = {}
            except TypeError:
                # Python 3.10 raises TypeError: issubclass() arg 1 must be a class
                # when calling issubclass(annotation, SheetList) with the dict annotation
                model_fields = {}
            headers_by_sheet[sheet_name] = [
                (field.alias or field_name) if by_alias else field_name
                for field_name, field in model_fields.items()
                if field_name != "validators_to_skip" and not field.exclude
            ]
        return headers_by_sheet

    @property
    def display_name(self):
        return uri_display_name(self.metadata.identifier)

    def dump(
        self,
        entities_exclude_defaults: bool = True,
        sort: bool = False,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx | None = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
    ) -> dict[str, Any]:
        """Dump the model to a dictionary.

        This is used in the Exporters to dump rules in the required format.

        Args:
            entities_exclude_defaults: Whether to exclude default prefix (and version) for entities.
                For example, given a class that is dumped as 'my_prefix:MyClass', if the prefix for the rules
                set in metadata.prefix = 'my_prefix', then this class will be dumped as 'MyClass' when this flag is set.
                Defaults to True.
            sort: Whether to sort the entities in the output.
            mode: The mode in which `to_python` should run.
                If mode is 'json', the output will only contain JSON serializable types.
                If mode is 'python', the output may contain non-JSON-serializable Python objects.
            by_alias: Whether to use the field's alias in the dictionary key if defined.
            exclude: A set of fields to exclude from the output.
            exclude_none: Whether to exclude fields that have a value of `None`.
            exclude_unset: Whether to exclude fields that have not been explicitly set.
            exclude_defaults: Whether to exclude fields that are set to their default value.
        """
        if sort:
            for field_name in self.model_fields.keys():
                value = getattr(self, field_name)
                # Ensure deterministic order of properties, classes, views, and so on
                if isinstance(value, SheetList):
                    value.sort(key=lambda x: x._identifier())

        context: dict[str, Any] = {}
        if entities_exclude_defaults:
            context["metadata"] = self.metadata

        exclude_input: IncEx | None = exclude

        return self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude_input,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            context=context,
        )


class SheetRow(SchemaModel):
    neatId: URIRefType | None = Field(
        alias="Neat ID",
        description="Globally unique identifier for the property",
        default=None,
    )

    @abstractmethod
    def _identifier(self) -> tuple[Hashable, ...]:
        raise NotImplementedError()

    def __repr__(self) -> str:
        # Simplified representation of the object for debugging
        return f"{self.__class__.__name__}({self._identifier()})"


T_SheetRow = TypeVar("T_SheetRow", bound=SheetRow)


class SheetList(list, MutableSequence[T_SheetRow]):
    @classmethod
    def __get_pydantic_core_schema__(cls, source: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        if args := get_args(source):
            item_type = args[0]
        else:
            # Someone use SheetList without specifying the type
            raise TypeError("SheetList must be used with a type argument, e.g., SheetList[InformationProperty]")

        instance_schema = core_schema.is_instance_schema(cls)
        sequence_row_schema = handler.generate_schema(Sequence[item_type])  # type: ignore[valid-type]

        non_instance_schema = core_schema.no_info_after_validator_function(SheetList, sequence_row_schema)
        return core_schema.union_schema([instance_schema, non_instance_schema])

    def to_pandas(self, drop_na_columns: bool = True, include: list[str] | None = None) -> pd.DataFrame:
        """Converts ResourceDict to pandas DataFrame."""
        df = pd.DataFrame([entity.model_dump() for entity in self])
        if drop_na_columns:
            df = df.dropna(axis=1, how="all")
        if include is not None:
            df = df[include]
        return df

    def _repr_html_(self) -> str:
        """Returns HTML representation of ResourceDict."""
        df = self.to_pandas(drop_na_columns=True)
        if "neatId" in df.columns:
            df = df.drop(columns=["neatId"])
        return df._repr_html_()  # type: ignore[operator]

    # Implemented to get correct type hints
    def __iter__(self) -> Iterator[T_SheetRow]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> T_SheetRow: ...

    @overload
    def __getitem__(self, index: slice) -> "SheetList[T_SheetRow]": ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> "T_SheetRow | SheetList[T_SheetRow]":
        if isinstance(index, slice):
            return SheetList[T_SheetRow](super().__getitem__(index))
        return super().__getitem__(index)


ExtensionCategoryType = Annotated[
    ExtensionCategory,
    PlainSerializer(
        lambda v: v.value if isinstance(v, ExtensionCategory) else v,
        return_type=str,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: ExtensionCategory(v) if isinstance(v, str) else v),
]


# Immutable such that this can be used as a key in a dictionary
class ContainerProperty(BaseModel, frozen=True):
    container: ContainerEntityType
    property_: DmsPropertyType


class ContainerDestinationProperty(ContainerProperty, frozen=True):
    value_type: DataType | ViewEntity
    connection: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None = None


class ViewRef(BaseModel, frozen=True):
    view: ViewEntityType


class ViewProperty(ViewRef, frozen=True):
    property_: DmsPropertyType



from dataclasses import dataclass

from cognite.neat._issues import NeatWarning


@dataclass(unsafe_hash=True)
class NeatValueWarning(NeatWarning):
    """{value}"""

    value: str


@dataclass(unsafe_hash=True)
class NotSupportedWarning(NeatWarning):
    """{feature} is not supported"""

    feature: str


@dataclass(unsafe_hash=True)
class RegexViolationWarning(NeatWarning):
    """The value '{value}' of {identifier} does not match the {pattern_name} pattern '{pattern}'"""

    extra = "{motivation}"
    value: str
    pattern: str
    identifier: str
    pattern_name: str
    motivation: str | None = None


@dataclass(unsafe_hash=True)
class MissingCogniteClientWarning(NeatWarning):
    """Missing Cognite Client required for {functionality}"""

    functionality: str



import re
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Literal

import pandas as pd
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import ContainerId, ViewId
from rdflib import Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules.models._base_input import InputComponent, InputRules
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import (
    ContainerEntity,
    DMSNodeEntity,
    DMSUnknownEntity,
    EdgeEntity,
    ReverseConnectionEntity,
    ViewEntity,
    load_connection,
    load_dms_value_type,
)
from cognite.neat._rules.models.entities._wrapped import DMSFilter
from cognite.neat._utils.rdf_ import uri_display_name

from ._rules import _DEFAULT_VERSION, DMSContainer, DMSEnum, DMSMetadata, DMSNode, DMSProperty, DMSRules, DMSView


@dataclass
class DMSInputMetadata(InputComponent[DMSMetadata]):
    space: str
    external_id: str
    creator: str
    version: str
    name: str | None = None
    description: str | None = None
    created: datetime | str | None = None
    updated: datetime | str | None = None
    logical: str | URIRef | None = None
    source_id: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DMSMetadata]:
        return DMSMetadata

    def dump(self) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        if self.created is None:
            output["created"] = datetime.now()
        if self.updated is None:
            output["updated"] = datetime.now()
        return output

    @classmethod
    def from_data_model(cls, data_model: dm.DataModelApply) -> "DMSInputMetadata":
        description, creator = cls._get_description_and_creator(data_model.description)
        return cls(
            space=data_model.space,
            name=data_model.name or None,
            description=description,
            external_id=data_model.external_id,
            version=data_model.version,
            creator=",".join(creator),
            created=datetime.now(),
            updated=datetime.now(),
        )

    @classmethod
    def _get_description_and_creator(cls, description_raw: str | None) -> tuple[str | None, list[str]]:
        if description_raw and (description_match := re.search(r"Creator: (.+)", description_raw)):
            creator = description_match.group(1).split(", ")
            description = description_raw.replace(description_match[0], "").strip() or None
        elif description_raw:
            creator = ["MISSING"]
            description = description_raw
        else:
            creator = ["MISSING"]
            description = None
        return description, creator

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    @property
    def identifier(self) -> URIRef:
        """Globally unique identifier for the data model.

        !!! note
            Unlike namespace, the identifier does not end with "/" or "#".

        """
        return DEFAULT_NAMESPACE[f"data-model/unverified/physical/{self.space}/{self.external_id}/{self.version}"]

    @property
    def namespace(self) -> Namespace:
        """Namespace for the data model used for the entities in the data model."""
        return Namespace(f"{self.identifier}/")


@dataclass
class DMSInputProperty(InputComponent[DMSProperty]):
    view: str
    view_property: str | None
    value_type: str | DataType | ViewEntity | DMSUnknownEntity
    name: str | None = None
    description: str | None = None
    connection: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | str | None = None
    nullable: bool | None = None
    immutable: bool | None = None
    is_list: bool | None = None
    default: str | int | dict | None = None
    container: str | None = None
    container_property: str | None = None
    index: str | list[str] | None = None
    constraint: str | list[str] | None = None
    neatId: str | URIRef | None = None
    logical: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DMSProperty]:
        return DMSProperty

    def dump(self, default_space: str, default_version: str) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        output["View"] = ViewEntity.load(self.view, space=default_space, version=default_version)
        output["Value Type"] = load_dms_value_type(self.value_type, default_space, default_version)
        output["Connection"] = load_connection(self.connection, default_space, default_version)
        output["Container"] = (
            ContainerEntity.load(self.container, space=default_space, version=default_version)
            if self.container
            else None
        )
        return output

    def referenced_view(self, default_space: str, default_version: str) -> ViewEntity:
        return ViewEntity.load(self.view, strict=True, space=default_space, version=default_version)

    def referenced_container(self, default_space: str) -> ContainerEntity | None:
        return ContainerEntity.load(self.container, strict=True, space=default_space) if self.container else None


@dataclass
class DMSInputContainer(InputComponent[DMSContainer]):
    container: str
    name: str | None = None
    description: str | None = None
    constraint: str | None = None
    neatId: str | URIRef | None = None
    used_for: Literal["node", "edge", "all"] | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DMSContainer]:
        return DMSContainer

    def dump(self, default_space: str) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        output["Container"] = self.as_entity_id(default_space)
        output["Constraint"] = (
            [ContainerEntity.load(constraint.strip(), space=default_space) for constraint in self.constraint.split(",")]
            if self.constraint
            else None
        )
        return output

    def as_entity_id(self, default_space: str) -> ContainerEntity:
        return ContainerEntity.load(self.container, strict=True, space=default_space)

    @classmethod
    def from_container(cls, container: dm.ContainerApply) -> "DMSInputContainer":
        constraints: list[str] = []
        for _, constraint_obj in (container.constraints or {}).items():
            if isinstance(constraint_obj, dm.RequiresConstraint):
                constraints.append(str(ContainerEntity.from_id(constraint_obj.require)))
            # UniquenessConstraint it handled in the properties
        container_entity = ContainerEntity.from_id(container.as_id())
        return cls(
            container=str(container_entity),
            name=container.name or None,
            description=container.description,
            constraint=", ".join(constraints) or None,
            used_for=container.used_for,
        )


@dataclass
class DMSInputView(InputComponent[DMSView]):
    view: str
    name: str | None = None
    description: str | None = None
    implements: str | None = None
    filter_: Literal["hasData", "nodeType", "rawFilter"] | str | None = None
    in_model: bool = True
    neatId: str | URIRef | None = None
    logical: str | URIRef | None = None

    def __post_init__(self):
        if self.in_model is None:
            self.in_model = True

    @classmethod
    def _get_verified_cls(cls) -> type[DMSView]:
        return DMSView

    def dump(self, default_space: str, default_version: str) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        output["View"] = self.as_entity_id(default_space, default_version)
        output["Implements"] = self._load_implements(default_space, default_version)
        return output

    def as_entity_id(self, default_space: str, default_version: str) -> ViewEntity:
        return ViewEntity.load(self.view, strict=True, space=default_space, version=default_version)

    def _load_implements(self, default_space: str, default_version: str) -> list[ViewEntity] | None:
        self.implements = self.implements.strip() if self.implements else None

        return (
            [
                ViewEntity.load(implement, strict=True, space=default_space, version=default_version)
                for implement in self.implements.split(",")
            ]
            if self.implements
            else None
        )

    def referenced_views(self, default_space: str, default_version: str) -> list[ViewEntity]:
        return self._load_implements(default_space, default_version) or []

    @classmethod
    def from_view(cls, view: dm.ViewApply, in_model: bool) -> "DMSInputView":
        view_entity = ViewEntity.from_id(view.as_id())

        return cls(
            view=str(view_entity),
            description=view.description,
            name=view.name,
            implements=", ".join([str(ViewEntity.from_id(parent, _DEFAULT_VERSION)) for parent in view.implements])
            or None,
            in_model=in_model,
            filter_=(str(DMSFilter.from_dms_filter(view.filter)) if view.filter else None),
        )


@dataclass
class DMSInputNode(InputComponent[DMSNode]):
    node: str
    usage: Literal["type", "collocation"]
    name: str | None = None
    description: str | None = None
    neatId: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DMSNode]:
        return DMSNode

    @classmethod
    def from_node_type(cls, node_type: dm.NodeApply) -> "DMSInputNode":
        return cls(node=f"{node_type.space}:{node_type.external_id}", usage="type")

    def dump(self, default_space: str, **_) -> dict[str, Any]:  # type: ignore[override]
        output = super().dump()
        output["Node"] = DMSNodeEntity.load(self.node, space=default_space)
        return output


@dataclass
class DMSInputEnum(InputComponent[DMSEnum]):
    collection: str
    value: str
    name: str | None = None
    description: str | None = None
    neatId: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DMSEnum]:
        return DMSEnum


@dataclass
class DMSInputRules(InputRules[DMSRules]):
    metadata: DMSInputMetadata
    properties: list[DMSInputProperty]
    views: list[DMSInputView]
    containers: list[DMSInputContainer] | None = None
    enum: list[DMSInputEnum] | None = None
    nodes: list[DMSInputNode] | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DMSRules]:
        return DMSRules

    def dump(self) -> dict[str, Any]:
        default_space = self.metadata.space
        default_version = str(self.metadata.version)

        return {
            "Metadata": self.metadata.dump(),
            "Properties": [prop.dump(default_space, default_version) for prop in self.properties],
            "Views": [view.dump(default_space, default_version) for view in self.views],
            "Containers": [container.dump(default_space) for container in self.containers or []] or None,
            "Enum": [enum.dump() for enum in self.enum or []] or None,
            "Nodes": [node_type.dump(default_space) for node_type in self.nodes or []] or None,
        }

    @classmethod
    def display_type_name(cls) -> str:
        return "UnverifiedDMSModel"

    @property
    def display_name(self):
        return uri_display_name(self.metadata.identifier)

    def _repr_html_(self) -> str:
        summary = {
            "type": "Physical Data Model",
            "intended for": "DMS Architect",
            "name": self.metadata.name,
            "space": self.metadata.space,
            "external_id": self.metadata.external_id,
            "version": self.metadata.version,
            "views": len(self.views),
            "containers": len(self.containers) if self.containers else 0,
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore

    @property
    def id_(self) -> URIRef:
        return DEFAULT_NAMESPACE[
            f"data-model/unverified/dms/{self.metadata.space}/{self.metadata.external_id}/{self.metadata.version}"
        ]

    def referenced_views_and_containers(self) -> tuple[set[ViewEntity], set[ContainerEntity]]:
        default_space = self.metadata.space
        default_version = self.metadata.version

        containers: set[ContainerEntity] = set()
        views = {parent for view in self.views for parent in view.referenced_views(default_space, default_version)}
        for prop in self.properties:
            views.add(prop.referenced_view(default_space, default_version))
            if ref_container := prop.referenced_container(default_space):
                containers.add(ref_container)

        return views, containers

    def as_view_entities(self) -> list[ViewEntity]:
        return [view.as_entity_id(self.metadata.space, self.metadata.version) for view in self.views]

    def as_container_entities(self) -> list[ContainerEntity]:
        return [container.as_entity_id(self.metadata.space) for container in self.containers or []]

    def imported_views_and_containers(self) -> tuple[set[ViewEntity], set[ContainerEntity]]:
        views, containers = self.referenced_views_and_containers()
        return views - set(self.as_view_entities()), containers - set(self.as_container_entities())

    def imported_views_and_containers_ids(self) -> tuple[set[ViewId], set[ContainerId]]:
        views, containers = self.imported_views_and_containers()
        return {view.as_id() for view in views}, {container.as_id() for container in containers}



import inspect
from abc import ABC, abstractmethod
from functools import lru_cache
from types import UnionType
from typing import Generic, TypeVar, Union, get_args, get_origin

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules._shared import ReadRules, Rules, VerifiedRules
from cognite.neat._rules.models import DMSInputRules, InformationInputRules
from cognite.neat._store._provenance import Agent as ProvenanceAgent

T_RulesIn = TypeVar("T_RulesIn", bound=Rules)
T_RulesOut = TypeVar("T_RulesOut", bound=Rules)
T_VerifiedIn = TypeVar("T_VerifiedIn", bound=VerifiedRules)
T_VerifiedOut = TypeVar("T_VerifiedOut", bound=VerifiedRules)


class RulesTransformer(ABC, Generic[T_RulesIn, T_RulesOut]):
    """This is the base class for all rule transformers."""

    @abstractmethod
    def transform(self, rules: T_RulesIn) -> T_RulesOut:
        """Transform the input rules into the output rules."""
        raise NotImplementedError()

    @property
    def agent(self) -> ProvenanceAgent:
        """Provenance agent for the importer."""
        return ProvenanceAgent(id_=DEFAULT_NAMESPACE[f"agent/{type(self).__name__}"])

    @property
    def description(self) -> str:
        """Get the description of the transformer."""
        return "MISSING DESCRIPTION"

    def is_valid_input(self, rules: T_RulesIn) -> bool:
        """Check if the input rules are valid."""
        types = self.transform_type_hint()
        for type_ in types:
            if get_origin(type_) is ReadRules:
                inner = get_args(type_)[0]
                if isinstance(rules, ReadRules) and isinstance(rules.rules, inner):
                    return True
            elif isinstance(rules, type_):
                return True
        return False

    @classmethod
    @lru_cache(maxsize=1)
    def transform_type_hint(cls) -> tuple[type, ...]:
        # This is an expensive operation, so we cache the result
        signature = inspect.signature(cls.transform)
        annotation = signature.parameters["rules"].annotation
        if isinstance(annotation, TypeVar):
            if annotation.__bound__ is None:
                raise TypeError(f"TypeVar {annotation} must be bound to a type.")
            annotation = annotation.__bound__
        # The annotation can be a type or a generic
        if get_origin(annotation) in [UnionType, Union]:
            return get_args(annotation)

        if get_origin(annotation) is ReadRules and isinstance(get_args(annotation)[0], TypeVar):
            # Hardcoded for now, as we only have two types of ReadRules
            return ReadRules[DMSInputRules], ReadRules[InformationInputRules]

        return (annotation,)


class VerifiedRulesTransformer(RulesTransformer[T_VerifiedIn, T_VerifiedOut], ABC): ...



import typing
import warnings
from collections import defaultdict
from collections.abc import Iterable, Set
from pathlib import Path
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes import Relationship, RelationshipList
from rdflib import Namespace, URIRef

from cognite.neat._issues.warnings import NeatValueWarning
from cognite.neat._shared import Triple
from cognite.neat._utils.auxiliary import create_sha256_hash

from ._base import DEFAULT_SKIP_METADATA_VALUES, ClassicCDFBaseExtractor, InstanceIdPrefix, T_CogniteResource


class RelationshipsExtractor(ClassicCDFBaseExtractor[Relationship]):
    """Extract data from Cognite Data Fusions Relationships into Neat."""

    _default_rdf_type = "Relationship"
    _instance_id_prefix = InstanceIdPrefix.relationship

    def __init__(
        self,
        items: Iterable[Relationship],
        namespace: Namespace | None = None,
        total: int | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
        prefix: str | None = None,
        identifier: typing.Literal["id", "externalId"] = "id",
    ):
        # This is used by the ClassicExtractor to log the target nodes, such
        # that it can extract them.
        # It is private to avoid exposing it to the user.
        self._target_external_ids_by_type: dict[InstanceIdPrefix, set[str]] = defaultdict(set)
        self._log_target_nodes = False
        # Ensure that this becomes an iterator, even if it is a list.
        to_iterate = (self._log_target_nodes_if_set(item) for item in items)
        super().__init__(
            to_iterate,
            namespace=namespace,
            total=total,
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
            camel_case=camel_case,
            as_write=as_write,
            prefix=prefix,
            identifier=identifier,
        )
        self._uri_by_external_id_by_type: dict[InstanceIdPrefix, dict[str, URIRef]] = defaultdict(dict)
        self._target_triples: list[tuple[URIRef, URIRef, str, str]] = []

    def _log_target_nodes_if_set(self, item: Relationship) -> Relationship:
        if not self._log_target_nodes:
            return item
        if item.target_type and item.target_external_id:
            self._target_external_ids_by_type[InstanceIdPrefix.from_str(item.target_type)].add(item.target_external_id)
        return item

    def _item2triples_special_cases(self, id_: URIRef, dumped: dict[str, Any]) -> list[Triple]:
        triples: list[Triple] = []
        if (source_external_id := dumped.pop("sourceExternalId")) and "sourceType" in dumped:
            source_type = dumped["sourceType"]
            source_prefix = InstanceIdPrefix.from_str(source_type)
            if self.identifier == "id":
                try:
                    source_uri = self._uri_by_external_id_by_type[source_prefix][source_external_id]
                except KeyError:
                    warnings.warn(
                        NeatValueWarning(f"Missing externalId {source_external_id} for {source_type}"), stacklevel=2
                    )
                else:
                    triples.append((id_, self.namespace["sourceExternalId"], source_uri))
            elif self.identifier == "externalId":
                triples.append(
                    (id_, self.namespace["sourceExternalId"], self.namespace[f"{source_prefix}{source_external_id}"])
                )
        if (target_external_id := dumped.pop("targetExternalId")) and "targetType" in dumped:
            target_type = dumped["targetType"]
            if self.identifier == "id":
                # We do not yet have the target nodes, so we log them for later extraction.
                self._target_triples.append((id_, self.namespace["targetExternalId"], target_type, target_external_id))
            elif self.identifier == "externalId":
                target_prefix = InstanceIdPrefix.from_str(target_type)
                triples.append(
                    (id_, self.namespace["targetExternalId"], self.namespace[f"{target_prefix}{target_external_id}"])
                )
        return triples

    @classmethod
    def _from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
    ) -> tuple[int | None, Iterable[Relationship]]:
        items = client.relationships(data_set_external_ids=data_set_external_id)
        return None, items

    @classmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError("Relationships do not have a hierarchy.")

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[Relationship]]:
        relationships = RelationshipList.load(Path(file_path).read_text())
        return len(relationships), relationships

    def _fallback_id(self, item: Relationship) -> str | None:
        if item.external_id:
            return create_sha256_hash(item.external_id)
        return None



"""This module performs importing of graph to TransformationRules pydantic class.
In more details, it traverses the graph and abstracts class and properties, basically
generating a list of rules based on which nodes that form the graph are made.
"""

from collections import UserDict, defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, cast

import pandas as pd
from cognite.client.utils._importing import local_import
from pandas import ExcelFile
from rdflib import Namespace, URIRef

from cognite.neat._issues import IssueList, MultiValueError
from cognite.neat._issues.errors import (
    FileMissingRequiredFieldError,
    FileNotFoundNeatError,
    FileReadError,
)
from cognite.neat._issues.warnings import FileMissingRequiredFieldWarning
from cognite.neat._rules._shared import ReadRules, T_InputRules
from cognite.neat._rules.models import (
    INPUT_RULES_BY_ROLE,
    VERIFIED_RULES_BY_ROLE,
    RoleTypes,
    SchemaCompleteness,
)
from cognite.neat._utils.spreadsheet import SpreadsheetRead, read_individual_sheet
from cognite.neat._utils.text import humanize_collection

from ._base import BaseImporter

SOURCE_SHEET__TARGET_FIELD__HEADERS = [
    (
        "Properties",
        "Properties",
        {
            RoleTypes.information: ["Class", "Property"],
            RoleTypes.dms: ["View", "View Property"],
        },
    ),
    ("Classes", "Classes", ["Class"]),
    ("Containers", "Containers", ["Container"]),
    ("Views", "Views", ["View"]),
    ("Enum", "Enum", ["Collection"]),
    ("Nodes", "Nodes", ["Node"]),
]


MANDATORY_SHEETS_BY_ROLE: dict[RoleTypes, set[str]] = {
    role_type: {
        str(sheet_name)
        for sheet_name in (
            VERIFIED_RULES_BY_ROLE.get(role_type).mandatory_fields(use_alias=True)  # type: ignore
            if VERIFIED_RULES_BY_ROLE.get(role_type)
            else []
        )
        if sheet_name is not None
    }
    for role_type in RoleTypes.__members__.values()
}


class MetadataRaw(UserDict):
    @classmethod
    def from_excel(cls, excel_file: ExcelFile, metadata_sheet_name: str) -> "MetadataRaw":
        return cls(pd.read_excel(excel_file, metadata_sheet_name, header=None).replace(float("nan"), None).values)

    @property
    def has_role_field(self) -> bool:
        return self.get("role") in [role.value for role in RoleTypes.__members__.values()]

    @property
    def role(self) -> RoleTypes:
        return RoleTypes(self["role"])

    @property
    def has_schema_field(self) -> bool:
        return self.get("schema") in [schema.value for schema in SchemaCompleteness.__members__.values()]

    @property
    def schema(self) -> SchemaCompleteness | None:
        if not self.has_schema_field:
            return None
        return SchemaCompleteness(self["schema"])

    def is_valid(self, issue_list: IssueList, filepath: Path) -> bool:
        if not self.has_role_field:
            issue_list.append(FileMissingRequiredFieldError(filepath, "metadata", "role"))
            return False

        return True


@dataclass
class ReadResult:
    sheets: dict[str, dict | list]
    read_info_by_sheet: dict[str, SpreadsheetRead]
    metadata: MetadataRaw
    prefixes: dict[str, Namespace] | None = None

    @property
    def role(self) -> RoleTypes:
        return self.metadata.role

    @property
    def schema(self) -> SchemaCompleteness | None:
        return self.metadata.schema


class SpreadsheetReader:
    def __init__(
        self,
        issue_list: IssueList,
        required: bool = True,
        metadata: MetadataRaw | None = None,
        sheet_prefix: Literal["", "Last", "Ref", "CDMRef"] = "",
    ):
        self.issue_list = issue_list
        self.required = required
        self.metadata = metadata
        self._sheet_prefix = sheet_prefix
        self._seen_files: set[Path] = set()
        self._seen_sheets: set[str] = set()

    @property
    def metadata_sheet_name(self) -> str:
        return f"{self._sheet_prefix}Metadata"

    @property
    def prefixes_sheet_name(self) -> str:
        return "Prefixes"

    @property
    def seen_sheets(self) -> set[str]:
        if not self._seen_files:
            raise ValueError("No files have been read yet.")
        return self._seen_sheets

    def sheet_names(self, role: RoleTypes) -> set[str]:
        names = MANDATORY_SHEETS_BY_ROLE[role]
        return {f"{self._sheet_prefix}{sheet_name}" for sheet_name in names if sheet_name != "Metadata"}

    def read(self, excel_file: pd.ExcelFile, filepath: Path) -> None | ReadResult:
        self._seen_files.add(filepath)
        self._seen_sheets.update(map(str, excel_file.sheet_names))
        metadata: MetadataRaw | None
        if self.metadata is not None:
            metadata = self.metadata
        else:
            metadata = self._read_metadata(excel_file, filepath)
            if metadata is None:
                # The reading of metadata failed, so we can't continue
                return None

        sheets, read_info_by_sheet = self._read_sheets(excel_file, metadata.role)
        if sheets is None or self.issue_list.has_errors:
            return None
        sheets["Metadata"] = dict(metadata)

        # Special case for reading prefixes as they are suppose to be read only once
        if (
            self.prefixes_sheet_name in excel_file.sheet_names
            and not self._sheet_prefix
            and (prefixes := self._read_prefixes(excel_file, filepath))
        ):
            sheets["Prefixes"] = prefixes

        return ReadResult(sheets, read_info_by_sheet, metadata)

    def _read_metadata(self, excel_file: ExcelFile, filepath: Path) -> MetadataRaw | None:
        if self.metadata_sheet_name not in excel_file.sheet_names:
            if self.required:
                self.issue_list.append(FileMissingRequiredFieldError(filepath, "sheet", self.metadata_sheet_name))
            return None

        metadata = MetadataRaw.from_excel(excel_file, self.metadata_sheet_name)

        if not metadata.is_valid(self.issue_list, filepath):
            return None
        return metadata

    def _read_prefixes(self, excel_file: ExcelFile, filepath: Path) -> dict[str, Namespace] | None:
        if self.prefixes_sheet_name not in excel_file.sheet_names:
            return None

        else:
            prefixes = {}

            for row in read_individual_sheet(excel_file, "Prefixes", expected_headers=["Prefix", "Namespace"]):
                if "Prefix" in row and "Namespace" in row:
                    prefixes[row["Prefix"]] = row["Namespace"]
                else:
                    if "Prefix" not in row:
                        self.issue_list.append(FileMissingRequiredFieldWarning(filepath, "prefixes", "prefix"))
                    if "Namespace" not in row:
                        self.issue_list.append(FileMissingRequiredFieldWarning(filepath, "prefixes", "namespace"))
                    return None

            return prefixes

    def _read_sheets(
        self, excel_file: ExcelFile, read_role: RoleTypes
    ) -> tuple[dict[str, dict | list] | None, dict[str, SpreadsheetRead]]:
        read_info_by_sheet: dict[str, SpreadsheetRead] = defaultdict(SpreadsheetRead)

        sheets: dict[str, dict | list] = {}

        expected_sheet_names = self.sheet_names(read_role)

        if missing_sheets := expected_sheet_names.difference(set(excel_file.sheet_names)):
            if self.required:
                self.issue_list.append(
                    FileMissingRequiredFieldError(
                        cast(Path, excel_file.io), "sheets", humanize_collection(missing_sheets)
                    )
                )
            return None, read_info_by_sheet

        for source_sheet_name, target_sheet_name, headers_input in SOURCE_SHEET__TARGET_FIELD__HEADERS:
            source_sheet_name = f"{self._sheet_prefix}{source_sheet_name}"

            if source_sheet_name not in excel_file.sheet_names:
                continue
            if isinstance(headers_input, dict):
                headers = headers_input[read_role]
            else:
                headers = headers_input

            try:
                sheets[target_sheet_name], read_info_by_sheet[source_sheet_name] = read_individual_sheet(
                    excel_file,
                    source_sheet_name,
                    return_read_info=True,
                    expected_headers=headers,
                )
            except Exception as e:
                self.issue_list.append(FileReadError(cast(Path, excel_file.io), str(e)))
                continue

        return sheets, read_info_by_sheet


class ExcelImporter(BaseImporter[T_InputRules]):
    """Import rules from an Excel file.

    Args:
        filepath (Path): The path to the Excel file.
    """

    def __init__(self, filepath: Path):
        self.filepath = filepath

    def to_rules(self) -> ReadRules[T_InputRules]:
        issue_list = IssueList(title=f"'{self.filepath.name}'")
        if not self.filepath.exists():
            raise FileNotFoundNeatError(self.filepath)

        with pd.ExcelFile(self.filepath) as excel_file:
            user_reader = SpreadsheetReader(issue_list)

            user_read = user_reader.read(excel_file, self.filepath)

        issue_list.trigger_warnings()
        if issue_list.has_errors:
            raise MultiValueError(issue_list.errors)

        if user_read is None:
            return ReadRules(None, {})

        sheets = user_read.sheets
        original_role = user_read.role
        read_info_by_sheet = user_read.read_info_by_sheet

        rules_cls = INPUT_RULES_BY_ROLE[original_role]
        rules = cast(T_InputRules, rules_cls.load(sheets))
        return ReadRules(rules, read_info_by_sheet)

    @property
    def description(self) -> str:
        return f"Excel file {self.filepath.name} read as unverified data model"

    @property
    def source_uri(self) -> URIRef:
        return URIRef(f"file://{self.filepath.name}")


class GoogleSheetImporter(BaseImporter[T_InputRules]):
    """Import rules from a Google Sheet.

    .. warning::

        This importer is experimental and may not work as expected.

    Args:
        sheet_id (str): The Google Sheet ID.
        skiprows (int): The number of rows to skip when reading the Google Sheet.
    """

    def __init__(self, sheet_id: str, skiprows: int = 1):
        self.sheet_id = sheet_id
        self.skiprows = skiprows

    def to_rules(self) -> ReadRules[T_InputRules]:
        raise NotImplementedError("Google Sheet Importer is not yet implemented.")

    def _get_sheets(self) -> dict[str, pd.DataFrame]:
        local_import("gspread", "google")
        import gspread  # type: ignore[import]

        client_google = gspread.service_account()
        google_sheet = client_google.open_by_key(self.sheet_id)
        return {worksheet.title: pd.DataFrame(worksheet.get_all_records()) for worksheet in google_sheet.worksheets()}
